/*! \file kf_based_terrain_analysis
 *
 *  \brief Library to analyze a 3D point cloud generating a point-level binary segmentation (obstacle / traversable)
 *  and a graph (roadmap) expressing the free space connectivity.
 *
 */

#ifndef _KF_BASED_TERRAIN_ANALYSIS_H
#define _KF_BASED_TERRAIN_ANALYSIS_H

#include <pcl/filters/passthrough.h>
#include <pcl/search/kdtree.h>
#include <pcl/filters/voxel_grid.h>

#include <iostream>
//#include <sstream>
//#include <iomanip>
//#include <memory>
#include <fstream>
#include <string>
//#include <sys/stat.h>
//#include <cmath>

#include "structs_definitions.h"
#include "time_supervisor.h"
#include "observation_downsampler.h"

#include "rapidcsv.h"
#include "shallow_neural_net.h"

class CKf_Based_Terrain_Analysis
{
private:
  CTimeSupervisor::Ptr performance_supervisor_ptr_;

  bool use_full_covariance_matrices_;
  std::vector<Eigen::Matrix3d> references_covariance_matrix_vector_;

  std::ofstream dataset_file_;
  bool extract_data_to_external_training_of_the_network_;
  bool use_neural_network_;
  bool first_iteration_;

  ShallowNeuralNet nn_;

  float predictZ(const pcl::PointXYZRGBNormal reference, const float delta_x, const float delta_y);

  float computeLikelihood(const pcl::PointXYZRGBNormal reference_in_sensor_frame,
                          const pcl::PointXYZRGBNormal point_in_sensor_frame);

  float mahalanobisDistance(const pcl::PointXYZRGBNormal reference, const pcl::PointXYZRGBNormal point_to_evaluate,
                            float &predicted_std_dev);

  float mahalanobisDistance(const pcl::PointXYZRGBNormal reference_in_sensor_frame,
                            const pcl::PointXYZRGBNormal point_in_sensor_frame, float &predicted_std_dev,
                            bool &observation_z_is_below_prediction);

  float mahalanobisDistance(const pcl::PointXYZRGBNormal reference_in_sensor_frame,
                            const pcl::PointXYZRGBNormal point_in_sensor_frame);

  float euclideanDistance(const pcl::PointXYZRGBNormal reference_in_sensor_frame,
                          const pcl::PointXYZRGBNormal point_in_sensor_frame);

  void estimateTangentPlane(
      const pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr observations_for_tangent_plane_estimation,
      const double measurement_std_dev, pcl::PointXYZRGBNormal &reference_in_sensor_frame);

  void generateRootVertex(const kf_based_terrain_analysis_lib::SensorConfiguration lidar_configuration,
                          const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration,
                          pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr ground_reference_cloud_ptr);

  void fastGenerateElevationCloud(const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration,
                                  pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr pcl_cloud_ptr,
                                  pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr elevation_cloud,
                                  std::vector<std::vector<int>> &correspondence_indexes);

  void labelGroundPoints(const kf_based_terrain_analysis_lib::SensorConfiguration lidar_configuration,
                         const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration,
                         const std::vector<std::vector<int> > &correspondence_indexes, const int reference_index,
                         pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr ground_reference_cloud_ptr,
                         pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr elevation_cloud_ptr,
                         pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr cloud_ptr, std::vector<std::vector<int>> &edges,
                         std::vector<std::vector<kf_based_terrain_analysis_lib::Edge>> &discarded_edges);

  void extractPointsInROI(const kf_based_terrain_analysis_lib::SensorConfiguration lidar_configuration,
                          const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration,
                          const pcl::PointXYZRGBNormal reference,
                          const pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr elevation_cloud_ptr,
                          pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr points_in_ROI);

  void labelLocalGroundPoints(const kf_based_terrain_analysis_lib::SensorConfiguration lidar_configuration,
                              const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration,
                              const std::vector<std::vector<int> > &correspondence_indexes, const int reference_index,
                              pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr points_in_ROI,
                              pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr references_in_ROI,
                              pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr ground_reference_cloud,
                              pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr elevation_cloud_ptr,
                              pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr cloud_ptr,
                              std::vector<std::vector<int>> &edges,
                              std::vector<std::vector<kf_based_terrain_analysis_lib::Edge>> &discarded_edges);

  void createDenseGroundCloud(const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration,
                              const pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr ground_reference_cloud_ptr,
                              const kf_based_terrain_analysis_lib::SensorConfiguration lidar_configuration,
                              pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr pcl_dense_cloud_ptr);

  void fastLabelPointcloudUsingGroundModel(
      const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration,
      const pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr elevation_cloud_ptr,
      const std::vector<std::vector<int> > &correspondence_indexes,
      const pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr ground_reference_cloud_ptr,
      const std::vector<std::vector<int>> &edges, pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr pcl_cloud_ptr);

  void ensureThatConnectionsAreReflectedInBothEnds(
      const pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr ground_reference_cloud_ptr,
      std::vector<std::vector<int>> edges);

  void createFileToDumpDataset(const std::string filename);

  void loadNeuralNetwork(const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration);

  void prepareFileInterface(const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration);

public:
  CKf_Based_Terrain_Analysis();
  ~CKf_Based_Terrain_Analysis();

  void groundSegmentation(const kf_based_terrain_analysis_lib::SensorConfiguration lidar_configuration,
                          const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration,
                          pcl::PointCloud<pcl::PointXYZRGBNormal> &pcl_cloud,
                          pcl::PointCloud<pcl::PointXYZRGBNormal> &ground_references_pcl_cloud,
                          std::vector<std::vector<int>> &edges,
                          std::vector<std::vector<kf_based_terrain_analysis_lib::Edge>> &discarded_edges,
                          pcl::PointCloud<pcl::PointXYZRGBNormal> &ground_dense_reconstruction_pcl_cloud);

  void showPerformanceStatistics(void);

  void predictZWithVariance(const pcl::PointXYZRGBNormal reference, const float delta_x, const float delta_y,
                            float &z_predicted, float &var_z_predicted);

};

#endif
