#include "kf_based_terrain_analysis.h"

CKf_Based_Terrain_Analysis::CKf_Based_Terrain_Analysis()
{
  performance_supervisor_ptr_ = boost::make_shared<CTimeSupervisor>();
  use_full_covariance_matrices_ = false; // TODO: Extract parameter!!
  first_iteration_ = true;
  std::cout << "Ground terrain analysis processor created!" << std::endl;
}

CKf_Based_Terrain_Analysis::~CKf_Based_Terrain_Analysis()
{
  if (extract_data_to_external_training_of_the_network_)
  {
    //std::cout << "Closing file!" << std::endl;
    dataset_file_.close();
  }
}

void CKf_Based_Terrain_Analysis::showPerformanceStatistics(void)
{
  performance_supervisor_ptr_->printStatistics();
}

void CKf_Based_Terrain_Analysis::extractPointsInROI(
    const kf_based_terrain_analysis_lib::SensorConfiguration lidar_configuration,
    const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration,
    const pcl::PointXYZRGBNormal reference, const pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr elevation_cloud_ptr,
    pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr points_in_ROI)
{
//  if (filtering_configuration.measure_performance)
//    CFunctionMonitor performance_monitor("extractPointsInROI", performance_supervisor_ptr_);

//std::cout << "Extracting points in ROI: reference.x = " << reference.x << "    reference.y = " << reference.y
//    << "    reference.z = " << reference.z << std::endl;

  float x_min, x_max, y_min, y_max, x_search_range, y_search_range;

  if (reference.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX] > 0.5)
  {
    x_search_range = filtering_configuration.ROI_delta_x_and_y;
    y_search_range = filtering_configuration.ROI_delta_x_and_y;
  }
  else
  {
    x_search_range = filtering_configuration.ROI_shadow_area;
    y_search_range = filtering_configuration.ROI_shadow_area;
  }

  x_min = reference.x - x_search_range;
  y_min = reference.y - y_search_range;
  x_max = reference.x + x_search_range;
  y_max = reference.y + y_search_range;

  //std::cout << "Using shadow area reference point! " << std::endl;
  //std::cout << "search_range = " << filtering_configuration.ROI_delta_x_and_y << std::endl;

  // Extracting lidar points
  pcl::PassThrough < pcl::PointXYZRGBNormal > pass;
  pass.setInputCloud(elevation_cloud_ptr);
  pass.setFilterFieldName("x");
  pass.setFilterLimits(x_min, x_max);
  pass.filter(*points_in_ROI);
  pass.setInputCloud(points_in_ROI);
  pass.setFilterFieldName("y");
  pass.setFilterLimits(y_min, y_max);
  pass.filter(*points_in_ROI);

  //std::cout << "X limits = " << x_min << " ; " << x_max << std::endl;
  //std::cout << "Y limits = " << y_min << " ; " << y_max << std::endl;
  //std::cout << "Z ground ref = " << reference.z << std::endl;

  //std::cout << "Total num of elevation pointcloud points in ROI = " << points_in_ROI->points.size() << std::endl;
}

float CKf_Based_Terrain_Analysis::euclideanDistance(const pcl::PointXYZRGBNormal reference_in_sensor_frame,
                                                    const pcl::PointXYZRGBNormal point_in_sensor_frame)
{
  // Expectation
  float roll = reference_in_sensor_frame.data_c[GRND_REF_DATA_C_1_ROLL];
  float pitch = reference_in_sensor_frame.data_c[GRND_REF_DATA_C_2_PITCH];

  float delta_x = point_in_sensor_frame.x - reference_in_sensor_frame.x;
  float delta_y = point_in_sensor_frame.y - reference_in_sensor_frame.y;

  float prediction = reference_in_sensor_frame.z + delta_x * pitch + delta_y * roll;

  // Innovation
  float observation = point_in_sensor_frame.z;
  float innovation = observation - prediction;

  return (innovation);
}

float CKf_Based_Terrain_Analysis::computeLikelihood(const pcl::PointXYZRGBNormal reference_in_sensor_frame,
                                                    const pcl::PointXYZRGBNormal point_in_sensor_frame)
{
  float z_pred = 0.0;
  float var_z_pred = 0.0;

  float delta_x = point_in_sensor_frame.x - reference_in_sensor_frame.x;
  float delta_y = point_in_sensor_frame.y - reference_in_sensor_frame.y;

  predictZWithVariance(reference_in_sensor_frame, delta_x, delta_y, z_pred, var_z_pred);

  float sigma = sqrt(var_z_pred);
  float regularization_term = 1.0 / (sigma * 2.506628275);

  float likelihood = regularization_term
      * (exp(-1.0 * ((point_in_sensor_frame.z - z_pred) * (point_in_sensor_frame.z - z_pred)) / (2 * var_z_pred)));
  //DEBUG!float likelihood = ((point_in_sensor_frame.z - z_pred) * (point_in_sensor_frame.z - z_pred)) / (2 * var_z_pred);

  if (false) //point_in_sensor_frame.z < -1.75)
  {
    std::cout << "delta_x = " << delta_x << std::endl;
    std::cout << "delta_y = " << delta_y << std::endl;

    std::cout << "z_pred = " << z_pred << std::endl;
    std::cout << "var_z_pred = " << var_z_pred << std::endl;

    std::cout << "sigma = " << sigma << std::endl;
    std::cout << "regularization_term = " << regularization_term << std::endl;

    std::cout << "z observed = " << point_in_sensor_frame.z << std::endl;
    std::cout << "likelihood = " << likelihood << std::endl;

    std::cout << "reference_in_sensor_frame.x = " << reference_in_sensor_frame.x << "    reference_in_sensor_frame.y = "
        << reference_in_sensor_frame.y << "    reference_in_sensor_frame.z = " << reference_in_sensor_frame.z
        << std::endl;

    getchar();
  }
  return (likelihood);
}

float CKf_Based_Terrain_Analysis::mahalanobisDistance(const pcl::PointXYZRGBNormal reference_in_sensor_frame,
                                                      const pcl::PointXYZRGBNormal point_in_sensor_frame,
                                                      float &predicted_std_dev)
{
  //std::cout << "<------- Exectuting CKf_Based_Terrain_Analysis::mahalanobisDistance... --------->" << std::endl;

  float prediction_fast = reference_in_sensor_frame.z
      + (point_in_sensor_frame.x - reference_in_sensor_frame.x)
          * reference_in_sensor_frame.data_c[GRND_REF_DATA_C_2_PITCH]
      + (point_in_sensor_frame.y - reference_in_sensor_frame.y)
          * reference_in_sensor_frame.data_c[GRND_REF_DATA_C_1_ROLL];

  float prediction_std_fast = sqrt(
      reference_in_sensor_frame.data_c[GRND_REF_DATA_C_3_Z_VARIANCE]
          + (point_in_sensor_frame.x - reference_in_sensor_frame.x)
              * (point_in_sensor_frame.x - reference_in_sensor_frame.x)
              * reference_in_sensor_frame.data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE]
          + (point_in_sensor_frame.y - reference_in_sensor_frame.y)
              * (point_in_sensor_frame.y - reference_in_sensor_frame.y)
              * reference_in_sensor_frame.data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE]);

  float mahalanobis_distance_fast = fabs(point_in_sensor_frame.z - prediction_fast) / prediction_std_fast;

  predicted_std_dev = prediction_std_fast;

  //assert(!std::isnan(mahalanobis_distance_fast));
  if (std::isnan(mahalanobis_distance_fast))
  {
    std::cout << "ref (x,y,z)   = " << reference_in_sensor_frame.x << ", " << reference_in_sensor_frame.y << ", "
        << reference_in_sensor_frame.z << std::endl;
    std::cout << "point (x,y,z) = " << point_in_sensor_frame.x << ", " << point_in_sensor_frame.y << ", "
        << point_in_sensor_frame.z << std::endl;
    std::cout << "ref (z, sz, a, sa, b, sb) = " << reference_in_sensor_frame.z << ", "
        << reference_in_sensor_frame.data_c[GRND_REF_DATA_C_3_Z_VARIANCE] << ", "
        << reference_in_sensor_frame.data_c[GRND_REF_DATA_C_2_PITCH] << ", "
        << reference_in_sensor_frame.data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE] << ", "
        << reference_in_sensor_frame.data_c[GRND_REF_DATA_C_1_ROLL] << ", "
        << reference_in_sensor_frame.data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE] << std::endl;
  }
  return (mahalanobis_distance_fast);
}

float CKf_Based_Terrain_Analysis::mahalanobisDistance(const pcl::PointXYZRGBNormal reference_in_sensor_frame,
                                                      const pcl::PointXYZRGBNormal point_in_sensor_frame)
{
  //std::cout << "<------- Exectuting CKf_Based_Terrain_Analysis::mahalanobisDistance... --------->" << std::endl;
  float prediction_fast = reference_in_sensor_frame.z
      + (point_in_sensor_frame.x - reference_in_sensor_frame.x)
          * reference_in_sensor_frame.data_c[GRND_REF_DATA_C_2_PITCH]
      + (point_in_sensor_frame.y - reference_in_sensor_frame.y)
          * reference_in_sensor_frame.data_c[GRND_REF_DATA_C_1_ROLL];

  float prediction_std_fast = -1.0;

  if (!use_full_covariance_matrices_)
  {
    prediction_std_fast = sqrt(
        reference_in_sensor_frame.data_c[GRND_REF_DATA_C_3_Z_VARIANCE]
            + (point_in_sensor_frame.x - reference_in_sensor_frame.x)
                * (point_in_sensor_frame.x - reference_in_sensor_frame.x)
                * reference_in_sensor_frame.data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE]
            + (point_in_sensor_frame.y - reference_in_sensor_frame.y)
                * (point_in_sensor_frame.y - reference_in_sensor_frame.y)
                * reference_in_sensor_frame.data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE]);
  }
  else
  {
    Eigen::Vector3d partial_derivatives;
    partial_derivatives(0) = 1.0;
    partial_derivatives(1) = (point_in_sensor_frame.x - reference_in_sensor_frame.x);
    partial_derivatives(2) = (point_in_sensor_frame.y - reference_in_sensor_frame.y);

    prediction_std_fast = sqrt(
        partial_derivatives.transpose()
            * references_covariance_matrix_vector_[reference_in_sensor_frame.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]]
            * partial_derivatives);
  }

  float mahalanobis_distance_fast = fabs(point_in_sensor_frame.z - prediction_fast) / prediction_std_fast;

  if (std::isnan(mahalanobis_distance_fast))
  {
    std::cout << "ref (x,y,z)   = " << reference_in_sensor_frame.x << ", " << reference_in_sensor_frame.y << ", "
        << reference_in_sensor_frame.z << std::endl;
    std::cout << "point (x,y,z) = " << point_in_sensor_frame.x << ", " << point_in_sensor_frame.y << ", "
        << point_in_sensor_frame.z << std::endl;
    std::cout << "ref (z, sz, a, sa, b, sb) = " << reference_in_sensor_frame.z << ", "
        << reference_in_sensor_frame.data_c[GRND_REF_DATA_C_3_Z_VARIANCE] << ", "
        << reference_in_sensor_frame.data_c[GRND_REF_DATA_C_2_PITCH] << ", "
        << reference_in_sensor_frame.data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE] << ", "
        << reference_in_sensor_frame.data_c[GRND_REF_DATA_C_1_ROLL] << ", "
        << reference_in_sensor_frame.data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE] << std::endl;
  }

  return (mahalanobis_distance_fast);
}

float CKf_Based_Terrain_Analysis::mahalanobisDistance(const pcl::PointXYZRGBNormal reference_in_sensor_frame,
                                                      const pcl::PointXYZRGBNormal point_in_sensor_frame,
                                                      float &predicted_std_dev, bool &observation_z_is_below_prediction)
{
  //std::cout << "<------- Exectuting CKf_Based_Terrain_Analysis::mahalanobisDistance... --------->" << std::endl;

  float prediction_fast = reference_in_sensor_frame.z
      + (point_in_sensor_frame.x - reference_in_sensor_frame.x)
          * reference_in_sensor_frame.data_c[GRND_REF_DATA_C_2_PITCH]
      + (point_in_sensor_frame.y - reference_in_sensor_frame.y)
          * reference_in_sensor_frame.data_c[GRND_REF_DATA_C_1_ROLL];

  float prediction_std_fast = sqrt(
      reference_in_sensor_frame.data_c[GRND_REF_DATA_C_3_Z_VARIANCE]
          + (point_in_sensor_frame.x - reference_in_sensor_frame.x)
              * (point_in_sensor_frame.x - reference_in_sensor_frame.x)
              * reference_in_sensor_frame.data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE]
          + (point_in_sensor_frame.y - reference_in_sensor_frame.y)
              * (point_in_sensor_frame.y - reference_in_sensor_frame.y)
              * reference_in_sensor_frame.data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE]);

  float mahalanobis_distance_fast = fabs(point_in_sensor_frame.z - prediction_fast) / prediction_std_fast;

  predicted_std_dev = prediction_std_fast;

  if (point_in_sensor_frame.z < prediction_fast)
    observation_z_is_below_prediction = true;
  else
    observation_z_is_below_prediction = false;

  //assert(!std::isnan(mahalanobis_distance_fast));
  if (std::isnan(mahalanobis_distance_fast))
  {
    std::cout << "ref (x,y,z)   = " << reference_in_sensor_frame.x << ", " << reference_in_sensor_frame.y << ", "
        << reference_in_sensor_frame.z << std::endl;
    std::cout << "point (x,y,z) = " << point_in_sensor_frame.x << ", " << point_in_sensor_frame.y << ", "
        << point_in_sensor_frame.z << std::endl;
    std::cout << "ref (z, sz, a, sa, b, sb) = " << reference_in_sensor_frame.z << ", "
        << reference_in_sensor_frame.data_c[GRND_REF_DATA_C_3_Z_VARIANCE] << ", "
        << reference_in_sensor_frame.data_c[GRND_REF_DATA_C_2_PITCH] << ", "
        << reference_in_sensor_frame.data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE] << ", "
        << reference_in_sensor_frame.data_c[GRND_REF_DATA_C_1_ROLL] << ", "
        << reference_in_sensor_frame.data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE] << std::endl;
  }
  return (mahalanobis_distance_fast);
}

float CKf_Based_Terrain_Analysis::predictZ(const pcl::PointXYZRGBNormal reference, const float delta_x,
                                           const float delta_y)
{
  float roll_slope = reference.data_c[GRND_REF_DATA_C_1_ROLL];
  float pitch_slope = reference.data_c[GRND_REF_DATA_C_2_PITCH];

  float predicted_z = reference.z + delta_x * pitch_slope + delta_y * roll_slope;
  return (predicted_z);
}

void CKf_Based_Terrain_Analysis::predictZWithVariance(const pcl::PointXYZRGBNormal reference, const float delta_x,
                                                      const float delta_y, float &z_predicted, float &var_z_predicted)
{
  float z_local_frame = reference.z;
  float roll_slope = reference.data_c[GRND_REF_DATA_C_1_ROLL];
  float pitch_slope = reference.data_c[GRND_REF_DATA_C_2_PITCH];

  z_predicted = z_local_frame + delta_x * pitch_slope + delta_y * roll_slope;

  Eigen::Vector3f partial_derivatives;
  partial_derivatives(0) = 1;
  partial_derivatives(1) = delta_x;
  partial_derivatives(2) = delta_y;

  Eigen::Matrix3f cov = Eigen::Matrix3f::Zero();
  cov(0, 0) = reference.data_c[GRND_REF_DATA_C_3_Z_VARIANCE];
  cov(1, 1) = reference.data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE];
  cov(2, 2) = reference.data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE];

  var_z_predicted = partial_derivatives.transpose() * cov * partial_derivatives;
}

void CKf_Based_Terrain_Analysis::estimateTangentPlane(
    const pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr observations_for_tangent_plane_estimation,
    const double measurement_std_dev, pcl::PointXYZRGBNormal &reference_in_sensor_frame)
{
  Eigen::Vector3d x;
  x << (double)reference_in_sensor_frame.z, (double)reference_in_sensor_frame.data_c[GRND_REF_DATA_C_2_PITCH], (double)reference_in_sensor_frame.data_c[GRND_REF_DATA_C_1_ROLL];

  Eigen::Matrix3d P;
  P << (double)reference_in_sensor_frame.data_c[GRND_REF_DATA_C_3_Z_VARIANCE], 0.0, 0.0, 0.0, (double)reference_in_sensor_frame.data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE], 0.0, 0.0, 0.0, (double)reference_in_sensor_frame.data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE];

  //std::cout << "x prior = " << std::endl << x << std::endl;
  //std::cout << "P prior = " << std::endl << P << std::endl;

  Eigen::RowVector3d H;
  Eigen::Vector3d K;

  double r = measurement_std_dev * measurement_std_dev;

  for (pcl::PointCloud<pcl::PointXYZRGBNormal>::iterator i = observations_for_tangent_plane_estimation->begin();
      i != observations_for_tangent_plane_estimation->end(); ++i)
  {
    H << (double)1.0, (i->x - (double)reference_in_sensor_frame.x), (i->y - (double)reference_in_sensor_frame.y);

    //std::cout << "H = " << std::endl << H << std::endl;

    K = P * H.transpose() / (H * P * H.transpose() + r);
    //std::cout << "K = " << std::endl << K << std::endl;

    x = x + K * (i->z - (H * x));
    P = P - K * H * P;

    //std::cout << "x posterior = " << std::endl << x << std::endl;
    //std::cout << "P posterior = " << std::endl << P << std::endl;
  }

  reference_in_sensor_frame.z = (float)x(0);
  reference_in_sensor_frame.data_c[GRND_REF_DATA_C_2_PITCH] = (float)x(1);
  reference_in_sensor_frame.data_c[GRND_REF_DATA_C_1_ROLL] = (float)x(2);

  reference_in_sensor_frame.data_c[GRND_REF_DATA_C_3_Z_VARIANCE] = (float)P(0, 0);
  reference_in_sensor_frame.data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE] = (float)P(1, 1);
  reference_in_sensor_frame.data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE] = (float)P(2, 2);

//  references_covariance_matrix_vector_[reference_in_sensor_frame.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]] = P;

  //std::cout << "x posterior = " << std::endl << x << std::endl;
  //std::cout << "P posterior = " << std::endl << P << std::endl;
  //getchar();
}

void CKf_Based_Terrain_Analysis::labelLocalGroundPoints(
    const kf_based_terrain_analysis_lib::SensorConfiguration lidar_configuration,
    const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration,
    const std::vector<std::vector<int> > &correspondence_indexes, const int reference_index,
    pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr points_in_ROI,
    pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr references_in_ROI,
    pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr ground_reference_cloud_ptr,
    pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr elevation_cloud_ptr,
    pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr cloud_ptr, std::vector<std::vector<int>> &edges,
    std::vector<std::vector<kf_based_terrain_analysis_lib::Edge>> &discarded_edges)
{
  // In this function we will check by Mahalanobis distance test (using the prior) each elevation_cloud point in the ROI of the reference being
  // evaluated. Those who pass, are ground points, and will be used to update the prior. Those who pass and that are not yet labeled (have
  // the class value as UNKNOWN)are also used to drive the exploration, being selected to create new references (vertices in the graph)
  //
  // Points classified as GROUND are changed to GROUND independently of their previous class (because points can be seen as obstacles from
  // a POV, but seen as ground from another). Points classified as OBSTACLE are changed to OBSTACLE only if they were not of GROUND class.
  // Either way, if a change of class is produced or if the new classification has a better score than the previous one, we mark the current reference
  // (vertex in the graph) as the one which made the label. This is done to avoid searching during the whole point cloud labeling.
  //
  // After this, we use the GROUND classified points in the elevation cloud to update the prior
  // and store the posterior in the graph
  //
  // Once we get the posterior, we use the points that were UNKNOWN at the beggining of this function to generate new references (vertices)
  // then, we analize the vertices in the ROI and make edges to those that are seen as ground from the currently processed vertex. The CONNECTED
  // _TO_ROOT flag is set in those to which we connect (it is unset by default)

  //std::cout << "Starting labelLocalGroundPoints!!" << std::endl;
//  if (filtering_configuration.measure_performance)
//    CFunctionMonitor performance_monitor("labelLocalGroundPoints", performance_supervisor_ptr_);

  CObservationDownsampler new_references_creator(filtering_configuration.ground_reference_search_resolution,
                                                 filtering_configuration.propagation_additive_noise,
                                                 filtering_configuration.propagation_z_additive_noise);

//  std::cout << "Reference_index = " << reference_index << std::endl;
//  std::cout << "Using reference with ID = "
//      << ground_reference_cloud_ptr->points[reference_index].data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX] << std::endl;
  pcl::PointXYZRGBNormal reference_in_sensor_frame = ground_reference_cloud_ptr->points[reference_index];

  pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr observations_for_tangent_plane_estimation(
      new pcl::PointCloud<pcl::PointXYZRGBNormal>);

// points in ROI come from elevation pointcloud, which is a reduction of the original
// pointcloud (we have stored the indexes to be able to label the original pointcloud)
//  std::cout << "Extracting observations for local tangent plane estimation..." << std::endl;
  for (pcl::PointCloud<pcl::PointXYZRGBNormal>::iterator i = points_in_ROI->begin(); i != points_in_ROI->end(); ++i)
  {
    pcl::PointXYZRGBNormal point_in_sensor_frame = *i;

    if ((int)std::floor(point_in_sensor_frame.data_c[DATA_C_1_ID_CLASS]) != OUTLIER)
    {
      float prediction_std_dev = -1.0;
//    std::cout << "Computing mahalanobis distance..." << std::endl;
      float mahalanobis_distance = mahalanobisDistance(reference_in_sensor_frame, point_in_sensor_frame,
                                                       prediction_std_dev);

//    std::cout << "Mahalanobis distance computed!" << std::endl;
      if (std::isnan(mahalanobis_distance))
      {
        std::cout << "prior pred sigma = " << prediction_std_dev << "    uncertainty = "
            << prediction_std_dev * filtering_configuration.mahalanobis_threshold << "    mahalanobis distance = "
            << mahalanobis_distance << std::endl;

        std::getchar();
      }

      if (mahalanobis_distance < filtering_configuration.mahalanobis_threshold)
      {
        //std::cout << "Ground point found! x = " << point_in_sensor_frame.x << "    y = " << point_in_sensor_frame.y
        //    << "    z = " << point_in_sensor_frame.z << std::endl << "mahalanobis_distance = " << mahalanobis_distance
        //    << "    mahalanobis_threshold = " << filtering_configuration.mahalanobis_threshold << std::endl;

//      std::cout << "Inserting observation for plane estimation!" << std::endl;
        observations_for_tangent_plane_estimation->push_back(point_in_sensor_frame); // independently of exploration, any ground point is
                                                                                     // used for plane estimation, to no discard useful
                                                                                     // information!

        if ((int)std::floor(point_in_sensor_frame.data_c[DATA_C_1_ID_CLASS]) == CLASS_UNKNOWN) // we use unknown points to drive the exploration
        {
//        std::cout << "Inserting observation to drive exploration! --> original index = "
//            << point_in_sensor_frame.data_c[DATA_C_3_ORIGINAL_INDEX] << std::endl;

//        std::cout << "elevation_cloud_ptr->points.size() = " << elevation_cloud_ptr->points.size() << std::endl;

          new_references_creator.addGroundObservationInSensorFrame(
              reference_in_sensor_frame,
              elevation_cloud_ptr->points[point_in_sensor_frame.data_c[DATA_C_3_ORIGINAL_INDEX]]);
        }

        // now we compute the score
        float score = 0.99 - (mahalanobis_distance / filtering_configuration.mahalanobis_threshold);
        if (score < 0.0)
          score = 0.0;

        float index_of_labelling_ref =
            elevation_cloud_ptr->points[point_in_sensor_frame.data_c[DATA_C_3_ORIGINAL_INDEX]].data_c[DATA_C_2_INDEX_OF_GROUND_REF_THAT_MADE_THE_LABEL];

//      std::cout << "Updating elevation cloud point to represent ground data" << std::endl;
        if ((int)std::floor(point_in_sensor_frame.data_c[DATA_C_1_ID_CLASS]) != CLASS_GROUND
            || std::floor((int)index_of_labelling_ref) == INDEX_UNKNOWN
            || index_of_labelling_ref - std::floor((int)index_of_labelling_ref) < score) // if the point was seen as obstacle from another POV
                                                                                         // or was unexplored, with set it to ground class
        {
          // and store the reference for the later full pointcloud labelling step (only if the score is better than the previous label
          elevation_cloud_ptr->points[point_in_sensor_frame.data_c[DATA_C_3_ORIGINAL_INDEX]].data_c[DATA_C_2_INDEX_OF_GROUND_REF_THAT_MADE_THE_LABEL] =
              reference_index + score;

          if ((int)std::floor(point_in_sensor_frame.data_c[DATA_C_1_ID_CLASS]) != CLASS_GROUND)
          {
            // set the class as ground
            elevation_cloud_ptr->points[point_in_sensor_frame.data_c[DATA_C_3_ORIGINAL_INDEX]].data_c[DATA_C_1_ID_CLASS] =
                (float)CLASS_GROUND + score;

            // colour the point properly
            elevation_cloud_ptr->points[point_in_sensor_frame.data_c[DATA_C_3_ORIGINAL_INDEX]].r = R_CLASS_GROUND;
            elevation_cloud_ptr->points[point_in_sensor_frame.data_c[DATA_C_3_ORIGINAL_INDEX]].g = G_CLASS_GROUND;
            elevation_cloud_ptr->points[point_in_sensor_frame.data_c[DATA_C_3_ORIGINAL_INDEX]].b = B_CLASS_GROUND;
          }
        }
      }
      else
      {
//      std::cout << "Obstacle point found! x = " << point_in_sensor_frame.x << "    y = " << point_in_sensor_frame.y
//          << "    z = " << point_in_sensor_frame.z << std::endl << "mahalanobis_distance = " << mahalanobis_distance
//          << "    mahalanobis_threshold = " << filtering_configuration.mahalanobis_threshold << std::endl;

        if ((int)std::floor(point_in_sensor_frame.data_c[DATA_C_1_ID_CLASS]) != CLASS_GROUND) // we don't want to convert ground points to obstacles
                                                                                              // because of the POV (one thing can be seem as obstacle or ground depending on the POV)
        {
          // Now we compute the score
          // ERROR!! Check this! it should be 1.0 - maha / thresh, isnt it?
          float score = (mahalanobis_distance / filtering_configuration.mahalanobis_threshold) - 1.0;
          if (score > 0.99)
            score = 0.99;

          // Now we check the score of the previous detection
          float index_of_labelling_ref =
              elevation_cloud_ptr->points[point_in_sensor_frame.data_c[DATA_C_3_ORIGINAL_INDEX]].data_c[DATA_C_2_INDEX_OF_GROUND_REF_THAT_MADE_THE_LABEL];

          if (std::floor((int)index_of_labelling_ref) == INDEX_UNKNOWN
              || index_of_labelling_ref - std::floor((int)index_of_labelling_ref) < score)
            elevation_cloud_ptr->points[point_in_sensor_frame.data_c[DATA_C_3_ORIGINAL_INDEX]].data_c[DATA_C_2_INDEX_OF_GROUND_REF_THAT_MADE_THE_LABEL] =
                reference_index + score;

          // set the class to OBSTACLE
          elevation_cloud_ptr->points[point_in_sensor_frame.data_c[DATA_C_3_ORIGINAL_INDEX]].data_c[DATA_C_1_ID_CLASS] =
              (float)CLASS_OBSTACLE + score; //0.99;

          // and colour the point
          elevation_cloud_ptr->points[point_in_sensor_frame.data_c[DATA_C_3_ORIGINAL_INDEX]].r = R_CLASS_OBSTACLE;
          elevation_cloud_ptr->points[point_in_sensor_frame.data_c[DATA_C_3_ORIGINAL_INDEX]].g = G_CLASS_OBSTACLE;
          elevation_cloud_ptr->points[point_in_sensor_frame.data_c[DATA_C_3_ORIGINAL_INDEX]].b = B_CLASS_OBSTACLE;
        }
      }
    }
  }
//  std::cout << "Observations for local tangent plane estimation extracted!" << std::endl;

  //std::cout << "Reference Prior = " << std::endl;
  //std::cout << "x = " << ground_reference_cloud_ptr->points[reference_index].x << "    y = "
  //    << ground_reference_cloud_ptr->points[reference_index].y << std::endl << "    z     = ( "
  //    << ground_reference_cloud_ptr->points[reference_index].z << " , "
  //    << sqrt(ground_reference_cloud_ptr->points[reference_index].data_c[GRND_REF_DATA_C_3_Z_VARIANCE]) << " )"
  //    << "    roll  = ( "
  //    << atan(ground_reference_cloud_ptr->points[reference_index].data_c[GRND_REF_DATA_C_1_ROLL]) * 180.0 / M_PI
  //    << " , "
  //    << sqrt(ground_reference_cloud_ptr->points[reference_index].data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE]) * 180.0
  //        / M_PI << " )" << "    pitch = ( "
  //    << atan(ground_reference_cloud_ptr->points[reference_index].data_c[GRND_REF_DATA_C_2_PITCH]) * 180.0 / M_PI
  //    << " , "
  //    << sqrt(ground_reference_cloud_ptr->points[reference_index].data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE]) * 180.0
  //        / M_PI << " )" << std::endl;

//  std::cout << "Estimating tangent plane..." << std::endl;
  estimateTangentPlane(observations_for_tangent_plane_estimation, (double)filtering_configuration.z_observation_std_dev,
                       reference_in_sensor_frame);
//  std::cout << "Posterior distribution obtained!" << std::endl;

  ground_reference_cloud_ptr->points[reference_index] = reference_in_sensor_frame;

//  std::cout << "Reference Posterior = " << std::endl;
//  std::cout << "x = " << ground_reference_cloud_ptr->points[reference_index].x << "    y = "
//      << ground_reference_cloud_ptr->points[reference_index].y << std::endl << "    z     = ( "
//      << ground_reference_cloud_ptr->points[reference_index].z << " , "
//      << sqrt(ground_reference_cloud_ptr->points[reference_index].data_c[GRND_REF_DATA_C_3_Z_VARIANCE]) << " )"
//      << "    roll  = ( "
//      << atan(ground_reference_cloud_ptr->points[reference_index].data_c[GRND_REF_DATA_C_1_ROLL]) * 180.0 / M_PI
//      << " , "
//      << sqrt(ground_reference_cloud_ptr->points[reference_index].data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE]) * 180.0
//          / M_PI << " )" << "    pitch = ( "
//      << atan(ground_reference_cloud_ptr->points[reference_index].data_c[GRND_REF_DATA_C_2_PITCH]) * 180.0 / M_PI
//      << " , "
//      << sqrt(ground_reference_cloud_ptr->points[reference_index].data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE]) * 180.0
//          / M_PI << " )" << std::endl;

//  std::cout << "Creating new references..." << std::endl;
  new_references_creator.generateNewGroundReferences(reference_index, reference_in_sensor_frame,
                                                     ground_reference_cloud_ptr, references_covariance_matrix_vector_,
                                                     edges);
//  std::cout << "New references created!" << std::endl;

//  std::cout << "Search for references in the ROI..." << std::endl;
  extractPointsInROI(lidar_configuration, filtering_configuration, ground_reference_cloud_ptr->points[reference_index],
                     ground_reference_cloud_ptr, references_in_ROI);
//  std::cout << "Search concluded! Number of references in ROI = " << references_in_ROI->size() << std::endl;

//  std::cout << "Making edges to existing references in the ROI... " << std::endl;
  int num_of_discarded_edges = 0;
  for (pcl::PointCloud<pcl::PointXYZRGBNormal>::iterator i = references_in_ROI->begin(); i != references_in_ROI->end();
      ++i)
  {
    pcl::PointXYZRGBNormal reference_in_ROI = *i;

    // just a check to not repeat already known edges
    if (reference_index != reference_in_ROI.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]
        && (std::find(edges[reference_index].begin(), edges[reference_index].end(),
                      reference_in_ROI.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]) == edges[reference_index].end())
        && (std::find(edges[reference_in_ROI.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]].begin(),
                      edges[reference_in_ROI.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]].end(), reference_index)
            == edges[reference_in_ROI.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]].end()))
    {
//      std::cout << "reference_index = " << reference_index << std::endl;
//      std::cout << "reference_in_ROI.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX] = "
//          << reference_in_ROI.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX] << std::endl;
//      std::cout << "edges[reference_index].size() = " << edges[reference_index].size() << std::endl;
//      //int n = 0;
//      //for (std::vector<int>::iterator i = edges[reference_index].begin(); i != edges[reference_index].end(); i++)
//      //{
//      //  std::cout << "edges[" << n++ << "] = " << *i << std::endl;
//      //}
//      std::cout << "Analizing vertex in ROI with id = " << reference_in_ROI.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]
//          << std::endl;

      // TODO: check if it would be convenient to use the whole cov matrix for this:
      float prediction_std_dev = -1.0;
      float mahalanobis_distance = mahalanobisDistance(reference_in_sensor_frame, reference_in_ROI, prediction_std_dev);

      // Just a check!
      if (std::isnan(mahalanobis_distance))
      {
        std::cout << "posterior pred sigma = " << prediction_std_dev << "    uncertainty = "
            << prediction_std_dev * filtering_configuration.mahalanobis_threshold << "    mahalanobis distance = "
            << mahalanobis_distance << std::endl;

        std::getchar();
      }
//      std::cout << "Checking connections..." << std::endl;
      bool connect = true;

      int edge_class = CLASS_EDGE_ACCEPTED;
      if (mahalanobis_distance > filtering_configuration.mahalanobis_threshold)
      {
        //std::cout << "Discarding connection! Mahalanobis distance = " << mahalanobis_distance
        //    << "    Mahalanobis threshold = " << filtering_configuration.mahalanobis_threshold << std::endl;
        connect = false;
        edge_class = CLASS_EDGE_DISCARDED_TOO_MUCH_MAHALANOBIS_DISTANCE;
      }

      if (connect)
      {
//        std::cout << "Adding edge between vertex " << reference_in_sensor_frame.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]
//            << " and vertex " << reference_in_ROI.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX] << std::endl;
        edges[reference_in_sensor_frame.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]].push_back(
            reference_in_ROI.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]);

        if (reference_in_sensor_frame.data_n[GRND_REF_DATA_N_2_VERTEX_CLASS] == VERTEX_CONNECTED_TO_ROOT)
          ground_reference_cloud_ptr->points[reference_in_ROI.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]].data_n[GRND_REF_DATA_N_2_VERTEX_CLASS] =
              VERTEX_CONNECTED_TO_ROOT;

      }
      else
      {
        kf_based_terrain_analysis_lib::Edge discarded_edge;
        discarded_edge.vertex_index = reference_in_ROI.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX];
        discarded_edge.edge_class = edge_class;
        discarded_edges[reference_in_sensor_frame.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]].push_back(discarded_edge);
        num_of_discarded_edges++;
      }
    }
    else
    {
//      std::cout << "Skip adding edge between vertex "
//          << reference_in_sensor_frame.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX] << " and vertex "
//          << reference_in_ROI.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX] << std::endl;
    }
  }
//  std::cout << "Total edges evaluated = " << references_in_ROI->points.size() << "    discarded = "
//      << num_of_discarded_edges << std::endl;
//getchar();
}

void CKf_Based_Terrain_Analysis::labelGroundPoints(
    const kf_based_terrain_analysis_lib::SensorConfiguration lidar_configuration,
    const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration,
    const std::vector<std::vector<int> > &correspondence_indexes, const int reference_index,
    pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr ground_reference_cloud_ptr,
    pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr elevation_cloud_ptr,
    pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr cloud_ptr, std::vector<std::vector<int>> &edges,
    std::vector<std::vector<kf_based_terrain_analysis_lib::Edge>> &discarded_edges)
{
  //if (filtering_configuration.measure_performance)
  //CFunctionMonitor performance_monitor("labelGroundPoints", performance_supervisor_ptr_);

  pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr points_in_ROI(new pcl::PointCloud<pcl::PointXYZRGBNormal>);
  pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr references_in_ROI(new pcl::PointCloud<pcl::PointXYZRGBNormal>);

  //std::cout << "Extracting points in ROI..." << std::endl;
  // We extract points from the elevation_cloud in the Region of Interest of the reference currently being processed
  extractPointsInROI(lidar_configuration, filtering_configuration, ground_reference_cloud_ptr->points[reference_index],
                     elevation_cloud_ptr, points_in_ROI);

//  std::cout << "Reference index = " << reference_index << std::endl;
//  std::cout << "Points in ROI extracted!! Number of points in ROI = " << points_in_ROI->points.size() << std::endl;
//  std::cout << "References in ROI extracted!! Number of references in ROI = " << references_in_ROI->points.size()
//      << std::endl;

  //std::cout << "Points in ROI extracted!" << std::endl;

//  std::cout << "Number of edges in reference prior to labeling = " << edges[reference_index].size() << std::endl;
  //std::cout << "Labeling local ground points..." << std::endl;
  labelLocalGroundPoints(lidar_configuration, filtering_configuration, correspondence_indexes, reference_index,
                         points_in_ROI, references_in_ROI, ground_reference_cloud_ptr, elevation_cloud_ptr, cloud_ptr,
                         edges, discarded_edges);
//  std::cout << "Local ground points labeled!" << std::endl;
//  std::cout << "Number of edges after labeling = " << edges[reference_index].size() << std::endl;

//  getchar();
}

void CKf_Based_Terrain_Analysis::fastGenerateElevationCloud(
    const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration,
    pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr pcl_cloud_ptr,
    pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr elevation_cloud, std::vector<std::vector<int>> &correspondence_indexes)
{
  // In this function we want to create a reduced point cloud using a grid. Each point in the reduced point cloud
  // will contain the lowest point of its grid cell, the values of z mean and variance, intensity mean and variance,
  // and a vector with the indices of the rest of points in the cell w.r.t. the original point cloud, also it will
  // be colored as OUTLIER or UNKNOWN (since ground obstacle classification is done later on the algorithm)

  //if (filtering_configuration.measure_performance)
  //CFunctionMonitor performance_monitor("generateElevationCloud", performance_supervisor_ptr_);

  // We will use a integer matrix with sufficient size to discretize the whole point cloud (2*MAX_RANGE/resolution)^2
  // This matrix is required to save memory, because it contains only one integer by cell, while the structure required
  // to store the actual data is far more complex (see the struct ElevationCloudCell), so we only create ElevationCloudCell
  // objects in cells that actually contain data (which are a lot less than the size of the grid)
  int number_of_rows_and_columns = std::ceil(2.0 * MAX_RANGE / filtering_configuration.elevation_grid_resolution);
  //std::cout << "number_of_rows_and_columns = " << number_of_rows_and_columns << std::endl;

  // We initialize the grid to unknown values
  int elevation_map[number_of_rows_and_columns][number_of_rows_and_columns];
  for (int i = 0; i < number_of_rows_and_columns; ++i)
    for (int j = 0; j < number_of_rows_and_columns; ++j)
      elevation_map[i][j] = INDEX_UNKNOWN;
  //std::cout << "elevation_map created! " << std::endl;

  std::vector<kf_based_terrain_analysis_lib::ElevationCloudCell> elevation_cell_vector;

  int i = 0;
  //std::cout << "Starting to fill the elevation_map! " << std::endl;
  for (pcl::PointCloud<pcl::PointXYZRGBNormal>::const_iterator it = pcl_cloud_ptr->begin(); it != pcl_cloud_ptr->end();
      ++it)
  {
    int row = std::round(
        (number_of_rows_and_columns / 2.0) + (it->x / filtering_configuration.elevation_grid_resolution));
    int col = std::round(
        (number_of_rows_and_columns / 2.0) + (it->y / filtering_configuration.elevation_grid_resolution));

    if (elevation_map[row][col] == INDEX_UNKNOWN)
    {
      elevation_map[row][col] = elevation_cell_vector.size();
      kf_based_terrain_analysis_lib::ElevationCloudCell new_cell;
      //std::cout << "Pointing to an empty cell, storing current point as lowest, and keeping its index! " << std::endl;
      //std::cout << "Point number " << i << "    row = " << row << "    col = " << col << std::endl;
      new_cell.lowest_point_in_cell = *it;
      new_cell.index_of_lowest_point = i;
      elevation_cell_vector.push_back(new_cell);
    }
    else
    {
      //std::cout << "Pointing to a NON-empty cell! checking z!" << std::endl;
      if (it->z < elevation_cell_vector[elevation_map[row][col]].lowest_point_in_cell.z)
      {
        //std::cout << "Current point is the lowest, storing it as it is!" << std::endl;
        elevation_cell_vector[elevation_map[row][col]].lowest_point_in_cell = *it;
        elevation_cell_vector[elevation_map[row][col]].index_of_lowest_point = i;
      }
    }
    //std::cout << "Storing z coordinate and index in the indexes vector" << std::endl;
    elevation_cell_vector[elevation_map[row][col]].z_coordinates.push_back(it->z);
    elevation_cell_vector[elevation_map[row][col]].intensities.push_back(it->data_n[DATA_N_0_INTENSITY]);
    //std::cout << "Storing intensity value: " << it->data_n[DATA_N_0_INTENSITY] << std::endl;
    elevation_cell_vector[elevation_map[row][col]].indexes_of_points_in_cell.push_back(i);
    i++;
  }

  //std::cout << "Now generating the elevation cloud from the elevation map!" << std::endl;
  //DEBUG
  //int total_number_of_points_stored = 0;

  for (int i = 0; i < elevation_cell_vector.size(); ++i)
  {
    pcl::PointXYZRGBNormal elevation_cloud_point = elevation_cell_vector[i].lowest_point_in_cell;
    if ((int)std::floor(elevation_cloud_point.data_c[DATA_C_1_ID_CLASS]) == OUTLIER)
    {
      elevation_cloud_point.r = R_CLASS_OUTLIER;
      elevation_cloud_point.g = G_CLASS_OUTLIER;
      elevation_cloud_point.b = B_CLASS_OUTLIER;
    }
    else
    {
      elevation_cloud_point.r = R_CLASS_UNKNOWN;
      elevation_cloud_point.g = G_CLASS_UNKNOWN;
      elevation_cloud_point.b = B_CLASS_UNKNOWN;
    }

    //std::cout << "Intensity of elevation point: mean = " << intensity_mean << "    std = " << sqrt(var_intensity) << std::endl;

    // Filling remaining fields
    elevation_cloud_point.data_n[DATA_N_0_ELEVATION_CLOUD_INDEX_OF_POINT_IN_ORIGINAL_CLOUD] =
        elevation_cell_vector[i].index_of_lowest_point;
    elevation_cloud_point.data_n[DATA_N_1_INTENSITY_VARIANCE] = 0.0;
    elevation_cloud_point.data_n[DATA_N_2_Z_MEAN] = 0.0;
    elevation_cloud_point.data_n[DATA_N_3_Z_VARIANCE] = 0.0;

    elevation_cloud_point.data_c[DATA_C_2_INDEX_OF_GROUND_REF_THAT_MADE_THE_LABEL] = INDEX_UNKNOWN;
    elevation_cloud_point.data_c[DATA_C_3_ORIGINAL_INDEX] = i; //storing the index for later in labelling step

    //std::cout << "Intensity of elevation point AFTER storing it: mean = " << elevation_cloud_point.data_n[DATA_N_0_INTENSITY] << "    std = " << sqrt( elevation_cloud_point.data_n[DATA_N_1_INTENSITY_VARIANCE]) << std::endl;

    //std::cout << "var z = " << it->data_n[DATA_N_3_Z_VARIANCE] << std::endl;
    //std::cout << "elevation_cell_vector[i].z_coordinates.size() = " << z_coordinates.size() << std::endl;

    //DEBUG!
    //total_number_of_points_stored += elevation_map[row][col].indexes_of_points_in_cell.size();

    correspondence_indexes.push_back(elevation_cell_vector[i].indexes_of_points_in_cell);
    //std::cout << "Storing point into elevation cloud!" << std::endl;
    elevation_cloud->push_back(elevation_cloud_point);
  }

  //std::cout << "elevation_cloud generated!!" << std::endl;
//  //std::cout << "total_number_of_points_stored = " << total_number_of_points_stored << std::endl;
//  //std::cout << "original number of points = " << pcl_cloud_ptr->points.size() << std::endl;
}

void CKf_Based_Terrain_Analysis::generateRootVertex(
    const kf_based_terrain_analysis_lib::SensorConfiguration lidar_configuration,
    const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration,
    pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr ground_reference_cloud_ptr)
{
  // In this function we generate the root vertex using the z value as -1.0 * LiDAR height, the slope values are set to 0.0
  // since we assume that the sensor is normal to the plane in which it lies. The uncertainties (sigma_z, sigma_roll, sigma_pitch)
  // are set by parameter (roll means not the angle, but the tangent of the angle, in other words the slope in y axis, and pitch
  // is the slope in x axis).
  // We also set the flag "VERTEX_CONNECTED_TO_ROOT" to true, because the root is obviously connected to root! --> this flag will be passed
  // through the graph building process in such a way that once it is built we can avoid to search paths to root using nodes that we already know
  // that are not connected with the root (the root position is same of the sensor but at ground level, so not connected to root means the same
  // as unreachable for the robot

//std::cout << "Generating root vertex" << std::endl;
// The first reference point (root vertex) is the ground in (0,0) approximately known by sensor extrinsic calibration
  pcl::PointXYZRGBNormal root_vertex;
  root_vertex.x = 0.0;
  root_vertex.y = 0.0;
  root_vertex.z = -1.0 * lidar_configuration.sensor_height;

  // We will keep the correspondence between ground references and original points in the lidar cloud to take advantage of the segmentation
  // for planning
  root_vertex.data_c[GRND_REF_DATA_C_0_RGB_CAST_INTO_FLOAT] = (float)INDEX_UNKNOWN;
//  root_vertex.r = 0;
//  root_vertex.g = 0;
//  root_vertex.b = 0;

  root_vertex.data_n[GRND_REF_DATA_N_2_VERTEX_CLASS] = VERTEX_CONNECTED_TO_ROOT;

  int index = 0;
  root_vertex.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX] = (float)index;

  // we set the prior
  root_vertex.data_c[GRND_REF_DATA_C_1_ROLL] = 0.0;
  root_vertex.data_c[GRND_REF_DATA_C_2_PITCH] = 0.0;
  root_vertex.data_c[GRND_REF_DATA_C_3_Z_VARIANCE] = filtering_configuration.z_initial_std_dev
      * filtering_configuration.z_initial_std_dev;
  root_vertex.data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE] = filtering_configuration.roll_initial_std_dev
      * filtering_configuration.roll_initial_std_dev;
  root_vertex.data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE] = filtering_configuration.pitch_initial_std_dev
      * filtering_configuration.pitch_initial_std_dev;

  root_vertex.data_n[GRND_REF_DATA_N_2_VERTEX_CLASS] = VERTEX_CONNECTED_TO_ROOT;
  ground_reference_cloud_ptr->points.push_back(root_vertex);

//  Eigen::Matrix3d covariance_matrix;
//  covariance_matrix << filtering_configuration.z_initial_std_dev * filtering_configuration.z_initial_std_dev, 0.0, 0.0, 0.0, filtering_configuration.pitch_initial_std_dev
//      * filtering_configuration.pitch_initial_std_dev, 0.0, 0.0, 0.0, filtering_configuration.roll_initial_std_dev
//      * filtering_configuration.roll_initial_std_dev;
//  references_covariance_matrix_vector_.push_back(covariance_matrix);
}

void CKf_Based_Terrain_Analysis::createDenseGroundCloud(
    const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration,
    const pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr ground_reference_cloud_ptr,
    const kf_based_terrain_analysis_lib::SensorConfiguration lidar_configuration,
    pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr pcl_dense_cloud_ptr)
{
  //if (filtering_configuration.measure_performance)
  CFunctionMonitor performance_monitor("createDenseGroundCloud", performance_supervisor_ptr_);

  //std::cout << "setting input cloud" << std::endl;
  pcl::KdTreeFLANN < pcl::PointXYZRGBNormal > kdtree;
  kdtree.setInputCloud(ground_reference_cloud_ptr);
  int K = 1;
  std::vector<int> pointIdxKNNSearch(K);
  std::vector<float> pointKNNSquaredDistance(K);

  //std::cout << "Starting dense cloud creation" << std::endl;
  float min_std_dev = 1000.0;
  float max_std_dev = -1.0;

  for (float i = -80.0; i < 80.0; i += 0.4)
  {
    for (float j = -80.0; j < 80.0; j += 0.4)
    {
      pcl::PointXYZRGBNormal point;
      point.x = i;
      point.y = j;

      kdtree.nearestKSearch(point, K, pointIdxKNNSearch, pointKNNSquaredDistance);
      const int NUM_OF_NEIGHBOURS = pointIdxKNNSearch.size();

      if (NUM_OF_NEIGHBOURS > 0)
      {
        //std::cout << "Computing Z value as a weighted average" << std::endl;
        float z_vector[NUM_OF_NEIGHBOURS];
        float z_var_vector[NUM_OF_NEIGHBOURS];
        for (int i = 0; i < NUM_OF_NEIGHBOURS; i++)
        {
          pcl::PointXYZRGBNormal reference = ground_reference_cloud_ptr->points[pointIdxKNNSearch[i]];

          float delta_x = point.x - reference.x;
          float delta_y = point.y - reference.y;
          predictZWithVariance(reference, delta_x, delta_y, z_vector[i], z_var_vector[i]);
        }

        float norm_factor = 0.0;
        float weights[NUM_OF_NEIGHBOURS];
        for (int i = 0; i < NUM_OF_NEIGHBOURS; i++)
        {
          if (z_var_vector[i] > 0.0)
          {
            weights[i] = 1.0 / z_var_vector[i];
            norm_factor += weights[i];
          }
          else
          {
            std::cout << "ERROR in CKf_Based_Terrain_Analysis::createDenseGroundCloud! --> invalid variance = "
                << z_var_vector[i] << std::endl;
            assert(false);
          }
        }

        for (int i = 0; i < NUM_OF_NEIGHBOURS; i++)
        {
          point.z += (weights[i] * z_vector[i]) / norm_factor;
        }

        //std::cout << "Computing color using std_dev: blue --> std_dev of zero; green -->std_dev >= MAX_STD_DEV
        float MAX_STD_DEV = filtering_configuration.max_pred_std_dev_for_labelling;
        float std_dev = sqrt(1.0 / norm_factor);

        if (std_dev < min_std_dev)
          min_std_dev = std_dev;

        if (std_dev > max_std_dev)
          max_std_dev = std_dev;

        //if (std_dev > MAX_STD_DEV)
        //  std_dev = MAX_STD_DEV;

        float red = 0.0;
        float green = 255.0 * std_dev / MAX_STD_DEV;
        float blue = 255.0 - red;

        point.r = (int)red;
        point.g = (int)green;
        point.b = (int)blue;

        if (std_dev < MAX_STD_DEV)
        {
          //std::cout << "Storing the point!" << std::endl;
          pcl_dense_cloud_ptr->push_back(point);
        }
      }
    }
  }
  //std::cout << "Dense reconstruction min_std_dev = " << min_std_dev << "  max_std_dev = " << max_std_dev << std::endl;
  //std::cout << "Finished dense cloud creation!!" << std::endl;
}

void CKf_Based_Terrain_Analysis::fastLabelPointcloudUsingGroundModel(
    const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration,
    const pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr elevation_cloud_ptr,
    const std::vector<std::vector<int> > &correspondence_indexes,
    const pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr ground_reference_cloud_ptr,
    const std::vector<std::vector<int>> &edges, pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr pcl_cloud_ptr)
{
  // In this function, we iterate over all points in the elevation cloud (the grid-based reduction of the original point cloud)
  // for each elevation cloud point, we extract the best reference (encoded in the integer part of the float in the field
  // DATA_C_2_INDEX_OF_GROUND_REF_THAT_MADE_THE_LABEL), and then using the indexes vector we check the mahalanobis distance to every
  // point represented by the elevation point and set the score.

  //if (filtering_configuration.measure_performance)
  //CFunctionMonitor performance_monitor("labelPointcloudUsingGroundModel", performance_supervisor_ptr_);
  //std::cout << "Labeling points in the original pointcloud using the posterior and the ROI indexes vectors..."
  //    << std::endl;

  for (pcl::PointCloud<pcl::PointXYZRGBNormal>::iterator i = elevation_cloud_ptr->begin();
      i != elevation_cloud_ptr->end(); ++i)
  {
    pcl::PointXYZRGBNormal elevation_cloud_point = *i;

    // We pick the reference that made the best prediction for this point in the elevation cloud
    int reference_index = (int)std::floor(
        elevation_cloud_point.data_c[DATA_C_2_INDEX_OF_GROUND_REF_THAT_MADE_THE_LABEL]);

    if (reference_index != INDEX_UNKNOWN)
    {
      // If this point in the elevation cloud has been labeled by a reference, we extract the reference to
      // check the Mahalanobis distance and make the actual labeling of the original point cloud
      pcl::PointXYZRGBNormal reference_in_sensor_frame = ground_reference_cloud_ptr->points[reference_index];

      // 1) We create a vectors to store the features to be used with the Shallow Neural Network for
      //    ground classification
      std::vector<int> features_indexes_in_pcl_cloud_ptr;

      std::vector<double> squared_distances;
      std::vector<double> incidence_angles;
      std::vector<double> intensities;
      std::vector<double> point_to_reference_squared_distances;
      std::vector<double> prediction_errors;
      std::vector<double> scores;
      std::vector<double> gt_label;

      // 2) We point to the vector of indices related to the elevation_cloud point under evaluation
      std::vector<std::vector<int>>::const_iterator index_iterator = correspondence_indexes.begin()
          + elevation_cloud_point.data_c[DATA_C_3_ORIGINAL_INDEX];

      // and iterate over this vector
      for (std::vector<int>::const_iterator point_iterator = index_iterator->begin();
          point_iterator != index_iterator->end(); ++point_iterator)
      {
        float mahalanobis_distance = mahalanobisDistance(reference_in_sensor_frame,
                                                         pcl_cloud_ptr->points[*point_iterator]);

        float score = 0.995 - (mahalanobis_distance / filtering_configuration.mahalanobis_threshold); // we use 0.995 instead of 1.0 to avoid overflowing the class field
        if (score < 0.0)
          score = 0.0;

        if (score > filtering_configuration.score_threshold) // We only use the Neural Network to segment ground points into several classes
        {
          if (!filtering_configuration.extract_data_to_external_training_of_the_network) // We dont want to change the ground truth when saving the dataset!
          {
            pcl_cloud_ptr->points[*point_iterator].r = R_CLASS_GROUND;
            pcl_cloud_ptr->points[*point_iterator].g = G_CLASS_GROUND;
            pcl_cloud_ptr->points[*point_iterator].b = B_CLASS_GROUND;
          }
          pcl::PointXYZRGBNormal point_to_evaluate = pcl_cloud_ptr->points[*point_iterator];

          if (filtering_configuration.use_neural_network
              || filtering_configuration.extract_data_to_external_training_of_the_network)
          {
            features_indexes_in_pcl_cloud_ptr.push_back(*point_iterator); // Index to modify the classification in a later step
            //std::cout << "Index saved = " << *point_iterator << std::endl;

            squared_distances.push_back(
                (double)(point_to_evaluate.x * point_to_evaluate.x + point_to_evaluate.y * point_to_evaluate.y
                    + point_to_evaluate.z * point_to_evaluate.z)); // Feature!

            double azimuth = atan2(point_to_evaluate.y, point_to_evaluate.x);
            double rho = sqrt(point_to_evaluate.x * point_to_evaluate.x + point_to_evaluate.y * point_to_evaluate.y);
            double slope_x = reference_in_sensor_frame.data_c[GRND_REF_DATA_C_2_PITCH];
            double slope_y = reference_in_sensor_frame.data_c[GRND_REF_DATA_C_1_ROLL];

            double incidence_angle = atan2(-1.0 * point_to_evaluate.z, rho)
                + atan(cos(azimuth) * slope_x + sin(azimuth) * slope_y);

            incidence_angles.push_back(incidence_angle); // Feature!
            intensities.push_back(point_to_evaluate.data_n[DATA_N_0_INTENSITY]); // Point intensity // Feature!

            double delta_x = (point_to_evaluate.x - reference_in_sensor_frame.x);
            double delta_y = (point_to_evaluate.y - reference_in_sensor_frame.y);
            double delta_z = (point_to_evaluate.z - reference_in_sensor_frame.z);

            point_to_reference_squared_distances.push_back(delta_x * delta_x + delta_y * delta_y + delta_z * delta_z); // Feature!

            prediction_errors.push_back((double)euclideanDistance(reference_in_sensor_frame, point_to_evaluate)); // Prediction Error // Feature!

            scores.push_back((double)score); // Feature!
          }

          pcl_cloud_ptr->points[*point_iterator].data_c[DATA_C_1_ID_CLASS] = (float)CLASS_GROUND + score;
        }
        else
        {
          if (euclideanDistance(reference_in_sensor_frame, pcl_cloud_ptr->points[*point_iterator])
              < filtering_configuration.robot_height)
          {
            if (!filtering_configuration.extract_data_to_external_training_of_the_network)
            {
              pcl_cloud_ptr->points[*point_iterator].r = R_CLASS_OBSTACLE;
              pcl_cloud_ptr->points[*point_iterator].g = G_CLASS_OBSTACLE;
              pcl_cloud_ptr->points[*point_iterator].b = B_CLASS_OBSTACLE;
            }
            // We don't use score with obstacle points, if they are considered obstacles, we believe that their
            // probability of being ground is zero, the class label is used to distinguish between general obstacles
            // and overhanging obstacles
            pcl_cloud_ptr->points[*point_iterator].data_c[DATA_C_1_ID_CLASS] = (float)CLASS_OBSTACLE + 0.0;
          }
          else
          {
            //std::cout << "Overhanging obstacle detected!" << std::endl;
            if (!filtering_configuration.extract_data_to_external_training_of_the_network)
            {
              pcl_cloud_ptr->points[*point_iterator].r = R_CLASS_OVERHANGING_OBSTACLE;
              pcl_cloud_ptr->points[*point_iterator].g = G_CLASS_OVERHANGING_OBSTACLE;
              pcl_cloud_ptr->points[*point_iterator].b = B_CLASS_OVERHANGING_OBSTACLE;
            }
            pcl_cloud_ptr->points[*point_iterator].data_c[DATA_C_1_ID_CLASS] = (float)CLASS_OVERHANGING_OBSTACLE + 0.0;
          }
        }
      }

      int num_of_ground_points = squared_distances.size();
      //std::cout << "num_of_ground_points = " << num_of_ground_points << std::endl;

      if (num_of_ground_points > 0
          && (filtering_configuration.use_neural_network
              || filtering_configuration.extract_data_to_external_training_of_the_network))
      {
        int total_num_of_points = correspondence_indexes[elevation_cloud_point.data_c[DATA_C_3_ORIGINAL_INDEX]].size();
        //std::cout << "total_num_of_points = " << total_num_of_points << std::endl;

        double ratio_ground_points = (double)num_of_ground_points / (double)total_num_of_points; // Feature!
        //std::cout << "ratio_ground_points = " << ratio_ground_points << std::endl;

        double sum = std::accumulate(intensities.begin(), intensities.end(), 0.0);
        double mean_intensity_ground_points_in_cell = sum / (double)num_of_ground_points; // Feature!
        //std::cout << "mean_intensity_ground_points_in_cell = " << mean_intensity_ground_points_in_cell << std::endl;

        double var_intensity_ground_points_in_cell = 0;
        for (int n = 0; n < num_of_ground_points; ++n)
          var_intensity_ground_points_in_cell += ((intensities[n] - mean_intensity_ground_points_in_cell)
              * (intensities[n] - mean_intensity_ground_points_in_cell));

        var_intensity_ground_points_in_cell /= (double)num_of_ground_points; // Feature!
        //std::cout << "var_intensity_ground_points_in_cell = " << var_intensity_ground_points_in_cell << std::endl;

        sum = std::accumulate(prediction_errors.begin(), prediction_errors.end(), 0.0);
        double mean_prediction_error = sum / (double)num_of_ground_points; // Feature!
        //std::cout << "mean_prediction_error = " << mean_prediction_error << std::endl;

        double var_prediction_error = 0;
        for (int n = 0; n < num_of_ground_points; ++n)
          var_prediction_error += ((prediction_errors[n] - mean_prediction_error)
              * (prediction_errors[n] - mean_prediction_error));

        var_prediction_error /= (double)num_of_ground_points; // Feature!
        //std::cout << "var_prediction_error = " << var_prediction_error << std::endl;

        sum = std::accumulate(scores.begin(), scores.end(), 0.0);
        double mean_of_scores = sum / (double)num_of_ground_points; // Feature!
        //std::cout << "mean_of_scores = " << mean_of_scores << std::endl;

        double var_of_scores = 0;
        for (int n = 0; n < num_of_ground_points; ++n)
          var_of_scores += ((scores[n] - mean_of_scores) * (scores[n] - mean_of_scores));

        var_of_scores /= (double)num_of_ground_points; // Feature!
        //std::cout << "var_of_scores = " << var_of_scores << std::endl;

        // This chunk of code is for generating a dataset to train the shallow neural network
        // We do this before predicting with the neural ne, so the ground truth is not altered
        if (filtering_configuration.extract_data_to_external_training_of_the_network)
        {
          int randomIndex = std::rand() % num_of_ground_points; // To save the dataset we will use only one random point of each elevation cloud
          //std::cout << "randomIndex = " << randomIndex << std::endl;

          int gt_label = -1;
          pcl::PointXYZRGBNormal point_to_read_gt_label =
              pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[randomIndex]];

          if (point_to_read_gt_label.r == R_CLASS_ROAD && point_to_read_gt_label.g == G_CLASS_ROAD
              && point_to_read_gt_label.b == B_CLASS_ROAD)
            gt_label = ROAD;

          if (point_to_read_gt_label.r == R_CLASS_TERRAIN && point_to_read_gt_label.g == G_CLASS_TERRAIN
              && point_to_read_gt_label.b == B_CLASS_TERRAIN)
            gt_label = TERRAIN;

          if (point_to_read_gt_label.r == R_CLASS_VEGETATION && point_to_read_gt_label.g == G_CLASS_VEGETATION
              && point_to_read_gt_label.b == B_CLASS_VEGETATION)
            gt_label = VEGETATION;

          if (point_to_read_gt_label.r == R_CLASS_SIDEWALK && point_to_read_gt_label.g == G_CLASS_SIDEWALK
              && point_to_read_gt_label.b == B_CLASS_SIDEWALK)
            gt_label = SIDEWALK;

          if (point_to_read_gt_label.r == R_CLASS_KITTI_OBSTACLE && point_to_read_gt_label.g == G_CLASS_KITTI_OBSTACLE
              && point_to_read_gt_label.b == B_CLASS_KITTI_OBSTACLE)
            gt_label = CLASS_OBSTACLE;

          if (point_to_read_gt_label.r == R_CLASS_OUTLIER && point_to_read_gt_label.g == G_CLASS_OUTLIER
              && point_to_read_gt_label.b == B_CLASS_OUTLIER)
            gt_label = OUTLIER;

          if (gt_label == -1)
          {
            std::cout << "gt_label = " << gt_label << std::endl;
            std::cout << "point_to_read_gt_label.r = " << (int)point_to_read_gt_label.r
                << "    point_to_read_gt_label.g = " << (int)point_to_read_gt_label.g
                << "    point_to_read_gt_label.b = " << (int)point_to_read_gt_label.b << std::endl;
          }

          assert(
              gt_label != -1
                  && "ERROR in CKf_Based_Terrain_Analysis::fastLabelPointcloudUsingGroundModel, unexpected class label!");

          if (gt_label != OUTLIER)
          {
            Eigen::VectorXd features(13, 1);

            features(0, 0) = squared_distances[randomIndex];
            features(1, 0) = incidence_angles[randomIndex];
            features(2, 0) = intensities[randomIndex];
            features(3, 0) = point_to_reference_squared_distances[randomIndex];
            features(4, 0) = prediction_errors[randomIndex];
            features(5, 0) = scores[randomIndex];
            features(6, 0) = ratio_ground_points;
            features(7, 0) = mean_intensity_ground_points_in_cell;
            features(8, 0) = var_intensity_ground_points_in_cell;
            features(9, 0) = mean_prediction_error;
            features(10, 0) = var_prediction_error;
            features(11, 0) = mean_of_scores;
            features(12, 0) = var_of_scores;

            //std::cout << "features = " << std::endl << features << std::endl;

            std::setiosflags(std::ios::fixed);
            std::setprecision(6);
            std::ostringstream s;

            s << features(0, 0) << ", " << features(1, 0) << ", " << features(2, 0) << ", " << features(3, 0) << ", "
                << features(4, 0) << ", " << features(5, 0) << ", " << features(6, 0) << ", " << features(7, 0) << ", "
                << features(8, 0) << ", " << features(9, 0) << ", " << features(10, 0) << ", " << features(11, 0)
                << ", " << features(12, 0) << ", " << gt_label << std::endl;

            if (features(0, 0) >= 0.0)
              dataset_file_ << s.str();

            //std::cout << "gt_label = " << gt_label << std::endl;
          }
        }

        // This chunk of code is for using the shallow neural net for traversability prediction
        if (filtering_configuration.use_neural_network
            && !filtering_configuration.extract_data_to_external_training_of_the_network)
        {
          for (int n = 0; n < num_of_ground_points; ++n)
          {
            Eigen::VectorXd features(nn_.number_of_features_, 1);

            features(0, 0) = squared_distances[n];
            features(1, 0) = incidence_angles[n];
            features(2, 0) = intensities[n];
            features(3, 0) = point_to_reference_squared_distances[n];
            features(4, 0) = prediction_errors[n];
            features(5, 0) = scores[n];
            features(6, 0) = ratio_ground_points;
            features(7, 0) = mean_intensity_ground_points_in_cell;
            features(8, 0) = var_intensity_ground_points_in_cell;
            features(9, 0) = mean_prediction_error;
            features(10, 0) = var_prediction_error;
            features(11, 0) = mean_of_scores;
            features(12, 0) = var_of_scores;

            int predicted_class = nn_.predict(features);

            switch (nn_.number_of_neurons_in_output_layer_)
            {
              case (2):
                if (predicted_class == CLASS_OBSTACLE)
                {
                  pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].data_c[DATA_C_1_ID_CLASS] =
                      (float)CLASS_OBSTACLE + 0.0;

                  pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].r = R_CLASS_NON_TRAVERSABLE;
                  pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].g = G_CLASS_NON_TRAVERSABLE;
                  pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].b = B_CLASS_NON_TRAVERSABLE;
                }
                else
                {
                  pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].data_c[DATA_C_1_ID_CLASS] =
                      (float)CLASS_GROUND + (float)scores[n];

                  pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].r = R_CLASS_GROUND;
                  pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].g = G_CLASS_GROUND;
                  pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].b = B_CLASS_GROUND;
                }
                //std::cout << "Class + score = " << pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].data_c[DATA_C_1_ID_CLASS] << std::endl;
                break;

              case (5):
                switch (predicted_class)
                {
                  case ROAD:
                    pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].r = R_CLASS_ROAD;
                    pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].g = G_CLASS_ROAD;
                    pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].b = B_CLASS_ROAD;
                    //std::cout << "NN says point is road!" << std::endl;
                    break;

                  case SIDEWALK:
                    pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].r = R_CLASS_SIDEWALK;
                    pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].g = G_CLASS_SIDEWALK;
                    pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].b = B_CLASS_SIDEWALK;
                    //std::cout << "NN says point is sidewalk!" << std::endl;
                    break;

                  case TERRAIN:
                    pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].r = R_CLASS_TERRAIN;
                    pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].g = G_CLASS_TERRAIN;
                    pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].b = B_CLASS_TERRAIN;
                    pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].data_c[DATA_C_1_ID_CLASS] =
                        (float)CLASS_OBSTACLE + 0.0;
                    //std::cout << "NN says point is terrain!" << std::endl;
                    break;

                  case VEGETATION:
                    pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].r = R_CLASS_VEGETATION;
                    pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].g = G_CLASS_VEGETATION;
                    pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].b = B_CLASS_VEGETATION;
                    pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].data_c[DATA_C_1_ID_CLASS] =
                        (float)CLASS_OBSTACLE + 0.0;
                    //std::cout << "NN says point is vegetation!" << std::endl;
                    break;

                  case CLASS_OBSTACLE:
                    pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].r = R_CLASS_KITTI_OBSTACLE;
                    pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].g = G_CLASS_KITTI_OBSTACLE;
                    pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].b = B_CLASS_KITTI_OBSTACLE;
                    pcl_cloud_ptr->points[features_indexes_in_pcl_cloud_ptr[n]].data_c[DATA_C_1_ID_CLASS] =
                        (float)CLASS_OBSTACLE + 0.0;
                    //std::cout << "NN says point is Obstacle!" << std::endl;
                    break;
                }
                break;

              default:
                assert(
                    "ERROR in CKf_Based_Terrain_Analysis::fastLabelPointcloudUsingGroundModel: Unrecognized number of outputs in Shallow Neural Network"
                        && false);
                break;
            }
          }
        }
      }
    }
    else // if we don't have enough information to try to predict the ground level at this point
    {
      if (!filtering_configuration.extract_data_to_external_training_of_the_network)
      {
        std::vector<std::vector<int>>::const_iterator index_iterator = correspondence_indexes.begin()
            + elevation_cloud_point.data_c[DATA_C_3_ORIGINAL_INDEX];

        for (std::vector<int>::const_iterator point_iterator = index_iterator->begin();
            point_iterator != index_iterator->end(); ++point_iterator)
        {
          // We now copy the data that is common to every point in the vector (because are statistics extracted
          // from the points in the vector)

          // in data_n[0] we have the point intensity, so nothing to change
          // in data_n[1] we will store the intensity variance in the elevation cloud cell
          pcl_cloud_ptr->points[*point_iterator].data_n[DATA_N_1_INTENSITY_VARIANCE] =
              elevation_cloud_point.data_n[DATA_N_1_INTENSITY_VARIANCE];

          // in data_n[2] we store the mean z value in the elevation cell
          pcl_cloud_ptr->points[*point_iterator].data_n[DATA_N_2_Z_MEAN] =
              elevation_cloud_point.data_n[DATA_N_2_Z_MEAN];

          // in data_n[3] we store the z variance in the elevation cell
          pcl_cloud_ptr->points[*point_iterator].data_n[DATA_N_3_Z_VARIANCE] =
              elevation_cloud_point.data_n[DATA_N_3_Z_VARIANCE];

          if (filtering_configuration.classify_not_labeled_points_as_obstacles)
          {
            if ((pcl_cloud_ptr->points[*point_iterator].z - elevation_cloud_point.z)
                < filtering_configuration.ground_threshold_in_not_analyzed_areas)
            {
              pcl_cloud_ptr->points[*point_iterator].r = R_CLASS_GROUND;
              pcl_cloud_ptr->points[*point_iterator].g = G_CLASS_GROUND;
              pcl_cloud_ptr->points[*point_iterator].b = B_CLASS_GROUND;

              float score = 0.0; // we don't have too much confidence in these points, because they are not analyzed
              pcl_cloud_ptr->points[*point_iterator].data_c[DATA_C_1_ID_CLASS] = (float)CLASS_GROUND + score;
            }
            else if ((pcl_cloud_ptr->points[*point_iterator].z - elevation_cloud_point.z)
                < filtering_configuration.robot_height)
            {
              pcl_cloud_ptr->points[*point_iterator].r = R_CLASS_OBSTACLE;
              pcl_cloud_ptr->points[*point_iterator].g = G_CLASS_OBSTACLE;
              pcl_cloud_ptr->points[*point_iterator].b = B_CLASS_OBSTACLE;

              float score = 0.0;
              pcl_cloud_ptr->points[*point_iterator].data_c[DATA_C_1_ID_CLASS] = (float)CLASS_OBSTACLE + score;
            }
            else
            {
              pcl_cloud_ptr->points[*point_iterator].r = R_CLASS_OVERHANGING_OBSTACLE;
              pcl_cloud_ptr->points[*point_iterator].g = G_CLASS_OVERHANGING_OBSTACLE;
              pcl_cloud_ptr->points[*point_iterator].b = B_CLASS_OVERHANGING_OBSTACLE;

              float score = 0.0;
              pcl_cloud_ptr->points[*point_iterator].data_c[DATA_C_1_ID_CLASS] = (float)CLASS_OVERHANGING_OBSTACLE
                  + score;
            }
          }
          else
          {
            pcl_cloud_ptr->points[*point_iterator].r = R_CLASS_UNKNOWN;
            pcl_cloud_ptr->points[*point_iterator].g = G_CLASS_UNKNOWN;
            pcl_cloud_ptr->points[*point_iterator].b = B_CLASS_UNKNOWN;

            float score = 0.0;
            pcl_cloud_ptr->points[*point_iterator].data_c[DATA_C_1_ID_CLASS] = (float)CLASS_UNKNOWN + score;
          }
        }
      }
    }
  }
  //std::cout << "Labeling concluded!" << std::endl;
}

void CKf_Based_Terrain_Analysis::ensureThatConnectionsAreReflectedInBothEnds(
    const pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr ground_reference_cloud_ptr, std::vector<std::vector<int>> edges)
{
  for (int vertex_iterator = 0; vertex_iterator < ground_reference_cloud_ptr->points.size(); ++vertex_iterator)
  {
//    if (ground_reference_cloud_ptr->points[vertex_iterator].data_n[GRND_REF_DATA_N_2_VERTEX_CLASS]
//        != VERTEX_NOT_CONNECTED_TO_ROOT)
//    {
    for (int connected_vertices_iterator = 0; connected_vertices_iterator < edges[vertex_iterator].size();
        ++connected_vertices_iterator)
    {
      int connected_vertex = edges[vertex_iterator][connected_vertices_iterator];
//        if (ground_reference_cloud_ptr->points[connected_vertex].data_n[GRND_REF_DATA_N_2_VERTEX_CLASS]
//            != VERTEX_NOT_CONNECTED_TO_ROOT)
//        {
      if (std::find(edges[connected_vertex].begin(), edges[connected_vertex].end(), vertex_iterator)
          == edges[connected_vertex].end())
        edges[connected_vertex].push_back(vertex_iterator);
//        }
    }
//    }
  }
}

void CKf_Based_Terrain_Analysis::loadNeuralNetwork(
    const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration)
{
  rapidcsv::Document doc(filtering_configuration.neural_net_filename_with_global_path.c_str(),
                         rapidcsv::LabelParams(-1, -1));

  nn_.number_of_features_ = filtering_configuration.neural_net_number_of_features;
  nn_.number_of_neurons_in_hidden_layer_ = filtering_configuration.neural_net_number_of_neurons_in_hidden_layer;
  nn_.number_of_neurons_in_output_layer_ = filtering_configuration.neural_net_number_of_neurons_in_output_layer;

  int b1_n_rows = nn_.number_of_neurons_in_hidden_layer_;
  int b1_n_cols = 1;
  nn_.b1_.resize(b1_n_rows, b1_n_cols);
  std::cout << "nn_.b1_.rows() = " << nn_.b1_.rows() << "    nn_.b1_.cols() = " << nn_.b1_.cols() << std::endl;

  int b1_row_ini = 0;
  int b1_row_fin = b1_row_ini + b1_n_rows - 1;

  int current_row_in_eigen = 0;
  for (int i = b1_row_ini; i <= b1_row_fin; ++i)
  {
    std::cout << "row index in file = " << i << std::endl;
    std::vector<double> row = doc.GetRow<double>(i);
    for (int j = 0; j < row.size(); ++j)
    {
      std::cout << "Writing value = " << row[j] << " in Eigen matrix at index ( " << current_row_in_eigen << ", " << j
          << " )" << std::endl;
      nn_.b1_(current_row_in_eigen, j) = row[j];
    }
    current_row_in_eigen++;
  }

  int b2_n_rows = nn_.number_of_neurons_in_output_layer_;
  int b2_n_cols = 1;
  nn_.b2_.resize(b2_n_rows, b2_n_cols);

  int b2_row_ini = b1_row_fin + 1;
  int b2_row_fin = b2_row_ini + b2_n_rows - 1;

  current_row_in_eigen = 0;
  for (int i = b2_row_ini; i <= b2_row_fin; ++i)
  {
    std::cout << "row index in file = " << i << std::endl;
    std::vector<double> row = doc.GetRow<double>(i);
    for (int j = 0; j < row.size(); ++j)
    {
      std::cout << "Writing value = " << row[j] << " in Eigen matrix at index ( " << current_row_in_eigen << ", " << j
          << " )" << std::endl;
      nn_.b2_(current_row_in_eigen, j) = row[j];
    }
    current_row_in_eigen++;
  }

  int gain_n_rows = nn_.number_of_features_;
  int gain_n_cols = 1;
  nn_.gain_.resize(gain_n_rows, gain_n_cols);

  int gain_row_ini = b2_row_fin + 1;
  int gain_row_fin = gain_row_ini + gain_n_rows - 1;

  current_row_in_eigen = 0;
  for (int i = gain_row_ini; i <= gain_row_fin; ++i)
  {
    std::vector<double> row = doc.GetRow<double>(i);
    for (int j = 0; j < row.size(); ++j)
      nn_.gain_(current_row_in_eigen, j) = row[j];

    current_row_in_eigen++;
  }

  int w1_n_rows = nn_.number_of_neurons_in_hidden_layer_;
  int w1_n_cols = nn_.number_of_features_;
  nn_.w1_.resize(w1_n_rows, w1_n_cols);

  int w1_row_ini = gain_row_fin + 1;
  int w1_row_fin = w1_row_ini + w1_n_rows - 1;

  current_row_in_eigen = 0;
  for (int i = w1_row_ini; i <= w1_row_fin; ++i)
  {
    std::vector<double> row = doc.GetRow<double>(i);
    for (int j = 0; j < row.size(); ++j)
      nn_.w1_(current_row_in_eigen, j) = row[j];

    current_row_in_eigen++;
  }

  int w2_n_rows = nn_.number_of_neurons_in_output_layer_;
  int w2_n_cols = nn_.number_of_neurons_in_hidden_layer_;
  nn_.w2_.resize(w2_n_rows, w2_n_cols);

  int w2_row_ini = w1_row_fin + 1;
  int w2_row_fin = w2_row_ini + w2_n_rows - 1;

  current_row_in_eigen = 0;
  for (int i = w2_row_ini; i <= w2_row_fin; ++i)
  {
    std::vector<double> row = doc.GetRow<double>(i);
    for (int j = 0; j < row.size(); ++j)
      nn_.w2_(current_row_in_eigen, j) = row[j];

    current_row_in_eigen++;
  }

  int x_offset_n_rows = nn_.number_of_features_;
  int x_offset_n_cols = 1;
  nn_.x_offset_.resize(x_offset_n_rows, x_offset_n_cols);

  int x_offset_row_ini = w2_row_fin + 1;
  int x_offset_row_fin = x_offset_row_ini + x_offset_n_rows - 1;

  current_row_in_eigen = 0;
  for (int i = x_offset_row_ini; i <= x_offset_row_fin; ++i)
  {
    std::vector<double> row = doc.GetRow<double>(i);
    for (int j = 0; j < row.size(); ++j)
      nn_.x_offset_(current_row_in_eigen, j) = row[j];

    current_row_in_eigen++;
  }

  int y_min_n_rows = 1;
  int y_min_n_cols = 1;
  nn_.y_min_.resize(y_min_n_rows, y_min_n_cols);

  int y_min_row_ini = x_offset_row_fin + 1;
  int y_min_row_fin = y_min_row_ini + y_min_n_rows - 1;

  current_row_in_eigen = 0;
  for (int i = y_min_row_ini; i <= y_min_row_fin; ++i)
  {
    std::vector<double> row = doc.GetRow<double>(i);
    for (int j = 0; j < row.size(); ++j)
      nn_.y_min_(current_row_in_eigen, j) = row[j];

    current_row_in_eigen++;
  }

  int total_rows_in_neural_network_description_file = b1_n_rows + b2_n_rows + gain_n_rows + w1_n_rows + w2_n_rows
      + x_offset_n_rows + y_min_n_rows;

  std::cout << "total_rows_in_neural_network_description_file = " << total_rows_in_neural_network_description_file
      << std::endl;

  std::cout << "b1 = " << std::endl << nn_.b1_ << std::endl;
  std::cout << "b2 = " << std::endl << nn_.b2_ << std::endl;
  std::cout << "gain = " << std::endl << nn_.gain_ << std::endl;
  std::cout << "w1 = " << std::endl << nn_.w1_ << std::endl;
  std::cout << "w2 = " << std::endl << nn_.w2_ << std::endl;
  std::cout << "x_offset = " << std::endl << nn_.x_offset_ << std::endl;
  std::cout << "y_min = " << std::endl << nn_.y_min_ << std::endl;

  //std::cout << "Waiting for the user to press any key..." << std::endl;
  //getchar();
  return;
}

void CKf_Based_Terrain_Analysis::createFileToDumpDataset(const std::string filename)
{
  std::cout << "Creating file to dump the dataset, already existing files with the same name will be erased! "
      << filename.c_str() << std::endl;

  std::cout << "Are you sure you want to continue? (y/n) " << std::endl;

  char response = 'n';
  response = getchar();

  if (response == 'y' || response == 'Y')
  {
    std::cout << "Creating file..." << std::endl;
    dataset_file_.open(filename.c_str(), std::ios_base::trunc);
    std::cout << "File created!" << std::endl;
  }
  else
  {
    std::cout << "File creation aborted!" << std::endl;
    assert(false && "Killing node!");
  }
  return;
}

void CKf_Based_Terrain_Analysis::prepareFileInterface(
    const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration)
{
  extract_data_to_external_training_of_the_network_ =
      filtering_configuration.extract_data_to_external_training_of_the_network;

  if (extract_data_to_external_training_of_the_network_)
    createFileToDumpDataset(filtering_configuration.dataset_filename_with_global_path);

  use_neural_network_ = filtering_configuration.use_neural_network;

  if (use_neural_network_)
    loadNeuralNetwork(filtering_configuration);
}

void CKf_Based_Terrain_Analysis::groundSegmentation(
    const kf_based_terrain_analysis_lib::SensorConfiguration lidar_configuration,
    const kf_based_terrain_analysis_lib::FilteringConfiguration filtering_configuration,
    pcl::PointCloud<pcl::PointXYZRGBNormal> &pcl_cloud,
    pcl::PointCloud<pcl::PointXYZRGBNormal> &ground_references_pcl_cloud, std::vector<std::vector<int>> &edges,
    std::vector<std::vector<kf_based_terrain_analysis_lib::Edge>> &discarded_edges,
    pcl::PointCloud<pcl::PointXYZRGBNormal> &ground_dense_reconstruction_pcl_cloud)
{
//  if (filtering_configuration.measure_performance)
//    CFunctionMonitor performance_monitor("groundSegmentation", performance_supervisor_ptr_);

  if (first_iteration_)
  {
    prepareFileInterface(filtering_configuration);
    first_iteration_ = false;
  }

//std::cout << "Starting ground segmentation!!" << std::endl;
//std::cout << "Initializing pointcloud: Number of points in pcl_cloud = " << pcl_cloud.points.size() << std::endl;

  // We copy the input in a pointer fashion (TODO: maybe it could be optimized)
  pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr pcl_cloud_ptr(new pcl::PointCloud<pcl::PointXYZRGBNormal>);
  *pcl_cloud_ptr = pcl_cloud;

// We create a new representation for the input cloud: Using a grid, we extract the lowest point of each cell
// as representative point of its area. Then we extract some statistics and using an indexes vector
// we keep track of all the points represented by each lowest point, so the original point cloud can be recovered
// at any time. The lowest points (those in elevation cloud) will be classified as being ground or obstacles using our
// prior knowledge and the ground points will be used to update the prior in a Kalman Filter fashion
  pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr elevation_cloud_ptr(new pcl::PointCloud<pcl::PointXYZRGBNormal>);
  std::vector < std::vector<int> > correspondence_indexes;

// The parameters estimated using the Kalman filter are the slopes and a z coordinate defining small local planes.
// Those planes along with their connectivities (if a plane sees other planes as ground are considered connected, otherwise
// (a plane sees other plane as obstacle) they are considered not-connected) are stored in a graph structure
// (ground_reference_cloud_ptr contains the vertices and the vector of int vectors named edges contains the edgesof the graph)
  pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr ground_reference_cloud_ptr(new pcl::PointCloud<pcl::PointXYZRGBNormal>);

// This is an optimized version of a previous function (thats why the "fast" is appended at the beginning of the name),
// but it became the standard so the previous one was deleted for sake of code clarity. This function generates the
// representation using the grid to keep only the lowest points in each cell and a vector with indexes to the rest of points
  fastGenerateElevationCloud(filtering_configuration, pcl_cloud_ptr, elevation_cloud_ptr, correspondence_indexes);
  //std::cout << "Elevation cloud generated! Npoints = " << elevation_cloud_ptr->points.size() << std::endl;

  // This function just creates the first vertex in the graph, using the prior information in the lidar and filtering configuration
  // structures.
  generateRootVertex(lidar_configuration, filtering_configuration, ground_reference_cloud_ptr);
  //std::cout << "Root vertex generated! Num of points = " << ground_reference_cloud_ptr->points.size() << std::endl;
  //    << "Number of edges = " << edges[0].size() << std::endl;

// We run the algorithm, starting from the reference at sensor origin it will try to find others trusty local
// ground references, which will increase the value of ground_reference_cloud->points.size(), evaluating
// local planes using these reference points, until no new references appear. The algorithm ends after evaluating
// all of them
  int reference_index = 0;
  while (reference_index < ground_reference_cloud_ptr->points.size())
  {
//    std::cout << "Evaluating reference num " << reference_index << ":" << std::endl << "x = "
//        << ground_reference_cloud_ptr->points[reference_index].x << "    y = "
//        << ground_reference_cloud_ptr->points[reference_index].y << "    z = "
//        << ground_reference_cloud_ptr->points[reference_index].z << std::endl;
////
//    std::cout << "r = " << (int)ground_reference_cloud_ptr->points[reference_index].r << "    g = "
//        << (int)ground_reference_cloud_ptr->points[reference_index].g << "    b = "
//        << (int)ground_reference_cloud_ptr->points[reference_index].b << std::endl;

    labelGroundPoints(lidar_configuration, filtering_configuration, correspondence_indexes, reference_index++,
                      ground_reference_cloud_ptr, elevation_cloud_ptr, pcl_cloud_ptr, edges, discarded_edges);

//    std::cout << "reference_index = " << reference_index << std::endl;
//    std::cout << "number of references = " << ground_reference_cloud_ptr->points.size() << std::endl;
  }
//  std::cout << "Total number of references = " << ground_reference_cloud_ptr->points.size() << std::endl;
//  std::cout << "Total number of covariance matrices = " << references_covariance_matrix_vector_.size() << std::endl;
//getchar();

  fastLabelPointcloudUsingGroundModel(filtering_configuration, elevation_cloud_ptr, correspondence_indexes,
                                      ground_reference_cloud_ptr, edges, pcl_cloud_ptr);

  for (int vertex_iterator = 0; vertex_iterator < ground_reference_cloud_ptr->points.size(); ++vertex_iterator)
  {
    // If the point in the original cloud has been classified as obstacle we will want to exclude it from the graph search that might be used for planning by other algorithms
    if (pcl_cloud_ptr->points[ground_reference_cloud_ptr->points[vertex_iterator].data_c[DATA_C_0_RGB_CAST_INTO_FLOAT]].data_c[DATA_C_1_ID_CLASS]
        == CLASS_OBSTACLE)
    {
      ground_reference_cloud_ptr->points[vertex_iterator].data_n[GRND_REF_DATA_N_2_VERTEX_CLASS] =
          VERTEX_NOT_CONNECTED_TO_ROOT;
      //edges[vertex_iterator].clear();
    }
  }

  // We can discard ground references if they are not connected with any other, to test this, we need
  // to be sure that connections are reflected in both ends
  ensureThatConnectionsAreReflectedInBothEnds(ground_reference_cloud_ptr, edges);

  // Dense reconstruction aims to serve as debugging or visualizing tool, it generates an XY regular grid with predicted
  // Z values using the closest reference point
  if (filtering_configuration.show_dense_reconstruction)
  {
//std::cout << "Ready to create dense reconstruction" << std::endl;
    pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr pcl_dense_cloud_ptr(new pcl::PointCloud<pcl::PointXYZRGBNormal>);
    createDenseGroundCloud(filtering_configuration, ground_reference_cloud_ptr, lidar_configuration,
                           pcl_dense_cloud_ptr);
    ground_dense_reconstruction_pcl_cloud.clear();
    ground_dense_reconstruction_pcl_cloud = *pcl_dense_cloud_ptr;
  }
  else
  {
    // TODO: Make this interface more clear
    ground_dense_reconstruction_pcl_cloud.clear();
    ground_dense_reconstruction_pcl_cloud = *elevation_cloud_ptr;
  }

  // (TODO: maybe it could be optimized)
  ground_references_pcl_cloud.clear();
  ground_references_pcl_cloud = *ground_reference_cloud_ptr;

  pcl_cloud.clear();
  pcl_cloud = *pcl_cloud_ptr;

//  references_covariance_matrix_vector_.clear();

  if (filtering_configuration.measure_performance)
    performance_supervisor_ptr_->printStatistics();
}

