/*
 * CObservationDownsampler.cpp
 *
 *  Created on: 2 Mar 2020
 *      Author: idelpino
 */

#include "../include/observation_downsampler.h"

CObservationDownsampler::CObservationDownsampler(float angular_resolution, float propagation_additive_noise,
                                                 float propagation_z_additive_noise)
{
  angular_resolution_ = angular_resolution;
  propagation_additive_noise_ = propagation_additive_noise;
  propagation_z_additive_noise_ = propagation_z_additive_noise;

  const int NUMBER_OF_REGIONS = round(2.0 * M_PI / angular_resolution_);
  //std::cout << "NUMBER_OF_REGIONS = " << NUMBER_OF_REGIONS << std::endl;

  observations_.reserve(NUMBER_OF_REGIONS);

  for (int i = 0; i < NUMBER_OF_REGIONS; ++i)
  {
    num_points_.push_back(0);
    std::vector<float> centroid;
    for (int j = 0; j < 2; ++j) // we need two simple accumulators to compute the mean of x and y
      centroid.push_back(0.0);

    centroids_.push_back(centroid);
    accumulator_set<float, stats<tag::mean, tag::variance> > accum_region; // but we want another more complex accumulator
                                                                           // to compute z variance
    acc_.push_back(accum_region);
    std::vector <pcl::PointXYZRGBNormal> default_pcl_point_vector;
    observations_.push_back(default_pcl_point_vector);
  }

  //std::cout << "CObservationDownsampler object created!!" << std::endl;
}

CObservationDownsampler::~CObservationDownsampler()
{
  // TODO Auto-generated destructor stub
}

void CObservationDownsampler::predictZWithVariance(const pcl::PointXYZRGBNormal reference, const float x, const float y,
                                                   float &z_predicted, float &var_z_predicted)
{
  z_predicted = reference.z + (x - reference.x) * reference.data_c[GRND_REF_DATA_C_2_PITCH]
      + (y - reference.y) * reference.data_c[GRND_REF_DATA_C_1_ROLL];

  Eigen::Vector3f partial_derivatives;
  partial_derivatives(0) = 1;
  partial_derivatives(1) = (y - reference.y);
  partial_derivatives(2) = (x - reference.x);

  Eigen::Matrix3f cov = Eigen::Matrix3f::Zero();
  cov(0, 0) = reference.data_c[GRND_REF_DATA_C_3_Z_VARIANCE];
  cov(1, 1) = reference.data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE];
  cov(2, 2) = reference.data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE];

  var_z_predicted = partial_derivatives.transpose() * cov * partial_derivatives;
}

void CObservationDownsampler::addGroundObservationInSensorFrame(
    const pcl::PointXYZRGBNormal reference_in_sensor_frame,
    const pcl::PointXYZRGBNormal ground_observation_in_sensor_frame)
{
  float obs_x_in_reference_frame = ground_observation_in_sensor_frame.x - reference_in_sensor_frame.x;
  float obs_y_in_reference_frame = ground_observation_in_sensor_frame.y - reference_in_sensor_frame.y;

  float azimuth_angle = atan2(obs_y_in_reference_frame, obs_x_in_reference_frame);
  if (azimuth_angle < 0.0)
    azimuth_angle += 2.0 * M_PI;

  if (azimuth_angle >= 2.0 * M_PI)
    azimuth_angle -= 2.0 * M_PI;

  int index = floor(azimuth_angle / angular_resolution_);

  // Just in case any point escapes the 0 to 2*PI wrapping
  const int NUMBER_OF_REGIONS = round(2.0 * M_PI / angular_resolution_);
  if (index == NUMBER_OF_REGIONS)
  {
    std::cout
        << "CObservationDownsampler::addGroundObservationInSensorFrame --> found a point escaping the angular wrapping!"
        << std::endl;
    std::cout << "azimuth_angle = " << azimuth_angle * 180.0 / M_PI << std::endl;
    std::cout << "angular_resolution = " << angular_resolution_ * 180.0 / M_PI << std::endl;
    std::cout << "INDEX circular exploration = " << index << std::endl;

    index = 0;
  }
  centroids_[index][0] += ground_observation_in_sensor_frame.x;
  centroids_[index][1] += ground_observation_in_sensor_frame.y;

  acc_[index](ground_observation_in_sensor_frame.z);

  num_points_[index]++;

  pcl::PointXYZRGBNormal aux = ground_observation_in_sensor_frame;
  // TODO: rename this to clarify the process
  aux.data_c[GRND_REF_DATA_C_3_Z_VARIANCE] = azimuth_angle;
  observations_[index].push_back(aux);
}

struct less_than_azimuth
{
    inline bool operator() (const pcl::PointXYZRGBNormal& p1, const pcl::PointXYZRGBNormal& p2)
    {
        return (p1.data_c[GRND_REF_DATA_C_3_Z_VARIANCE] < p2.data_c[GRND_REF_DATA_C_3_Z_VARIANCE]);
    }
};

pcl::PointXYZRGBNormal median(std::vector<pcl::PointXYZRGBNormal> observations)
{
  size_t size = observations.size();

  assert(size > 0 && "Error computing median (observation_downsampler.cpp), size = 0");

  sort(observations.begin(), observations.end(), less_than_azimuth());
  return observations[size / 2];
}

void CObservationDownsampler::generateNewGroundReferences(
    const int reference_index, const pcl::PointXYZRGBNormal &reference,
    pcl::PointCloud<pcl::PointXYZRGBNormal>::Ptr ground_reference_cloud_ptr,
    std::vector<Eigen::Matrix3d> &references_covariance_matrix_vector, std::vector<std::vector<int>> &edges)
{
  //std::cout << "Generating new vertices!" << std::endl;
  const int NUMBER_OF_REGIONS = floor(2.0 * M_PI / angular_resolution_);

  for (int i = 0; i < NUMBER_OF_REGIONS; ++i)
  {
    if (num_points_[i] != 0)
    {
      pcl::PointXYZRGBNormal point;
//      point.x = centroids_[i][0] /= (float)num_points_[i];
//      point.y = centroids_[i][1] /= (float)num_points_[i];
//
//      float z_centroid = mean(acc_[i]);
//      point.z = z_centroid;

      pcl::PointXYZRGBNormal median_point = median(observations_[i]);
      point.x = median_point.x;
      point.y = median_point.y;
      point.z = median_point.z;

      //std::cout << "Region = " << i << "    Num points = " << num_points_[i] << std::endl
      //    << "Centroid old version x, y, z = " << centroids_[i][0] << "    " << centroids_[i][1] << "    " << centroids_[i][2]
      //    << "    Centroid with accums = " << z_centroid << "    variance = " << z_centroid_variance << std::endl;
      //assert(z_centroid == point.z && "something goes wrong!");

      point.data_c[GRND_REF_DATA_C_0_RGB_CAST_INTO_FLOAT] = (float)median_point.data_n[DATA_N_0_ELEVATION_CLOUD_INDEX_OF_POINT_IN_ORIGINAL_CLOUD];
//      point.r = 0; // just to not leave uninitialized values
//      point.g = 0;
//      point.b = 0;

      // Warning! the index is computed as size() because we haven't added the point yet!
      point.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX] = (float)ground_reference_cloud_ptr->points.size();

      // we set the prior
      point.data_c[GRND_REF_DATA_C_1_ROLL] = reference.data_c[GRND_REF_DATA_C_1_ROLL];
      point.data_c[GRND_REF_DATA_C_2_PITCH] = reference.data_c[GRND_REF_DATA_C_2_PITCH];

      float range = sqrt(
          (point.x - reference.x) * (point.x - reference.x) + (point.y - reference.y) * (point.y - reference.y)
              + (point.z - reference.z) * (point.z - reference.z));

      float z_centroid_variance = variance(acc_[i]);

      float z_aux = 0.0;
      float z_aux_var = 0.0;
      predictZWithVariance(reference, point.x, point.y, z_aux, z_aux_var);
      z_aux_var += (propagation_z_additive_noise_ * range) * (propagation_z_additive_noise_ * range);

      float greatest_z_var = z_aux_var;
      if (greatest_z_var < z_centroid_variance)
        greatest_z_var = z_centroid_variance;

      point.data_c[GRND_REF_DATA_C_3_Z_VARIANCE] = greatest_z_var;

      //std::cout << "z std_dev from propagation = " << sqrt(z_aux_var) << ";   from ROI data = " << sqrt(z_centroid_variance) << std::endl;

      float q = (range * propagation_additive_noise_) * (range * propagation_additive_noise_);

      point.data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE] = reference.data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE] + q;

      point.data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE] = reference.data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE] + q;

      point.data_n[GRND_REF_DATA_N_2_VERTEX_CLASS] = VERTEX_NOT_CONNECTED_TO_ROOT; // we create it as not connected, and only if other vertex connected to root
                                                                                   // makes an edge to this one, then we change this value setting it to "connected"

      //std::cout << "Base reference roll std = " << sqrt(reference.data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE]) * 180.0 / M_PI << std::endl
      //<< "Base reference pitch std = " << sqrt(reference.data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE]) * 180.0 / M_PI
      //<< std::endl << "New Reference: x y z = " << point.x << " " << point.y << " " << point.z << std::endl
      //<< "New reference roll std = " << sqrt(point.data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE]) * 180.0 / M_PI << std::endl
      //<< "New reference pitch std = " << sqrt(point.data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE]) * 180.0 / M_PI << std::endl
      //<< std::endl << "Propagation additive noise std = " << (range * propagation_additive_noise_) * 180.0 / M_PI << std::endl;

      //getchar();

      // and store the point
      ground_reference_cloud_ptr->points.push_back(point);

//      //TODO: check this!
//      Eigen::Matrix3d covariance_matrix = references_covariance_matrix_vector[reference.data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]];
//      covariance_matrix(0,0) = point.data_c[GRND_REF_DATA_C_3_Z_VARIANCE];
//      covariance_matrix(1,1) = point.data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE];
//      covariance_matrix(2,2) = point.data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE];
//
////      covariance_matrix << point.data_c[GRND_REF_DATA_C_3_Z_VARIANCE], 0.0, 0.0, 0.0, point.data_n[GRND_REF_DATA_N_1_PITCH_VARIANCE], 0.0, 0.0, 0.0, point.data_n[GRND_REF_DATA_N_0_ROLL_VARIANCE];
//
//      references_covariance_matrix_vector.push_back(covariance_matrix);
//
//      assert(
//          ground_reference_cloud_ptr->points[ground_reference_cloud_ptr->points.size() - 1].data_n[GRND_REF_DATA_N_3_ORIGINAL_INDEX]
//              == references_covariance_matrix_vector.size() - 1
//              && "ERROR!! unxepected index in CObservationDownsampler::generateNewGroundReferences");
    }
  }

}

